import re
import warnings
from collections import OrderedDict
from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, Final, List, Optional, Set, Type, cast

import black
import black.parsing
import boto3
import botocore.model
import botocore.session
import isort
from docformatter.format import Formatter

from .docstring import DocumentationFormatter, FormatterArgs
from .methods import (  # pylint: disable=import-error
    CreateMethodGenerator,
    DeleteMethodGenerator,
    GeneralMethodGenerator,
    GetManyMethodGenerator,
    GetMethodGenerator,
    ListMethodGenerator,
    ManagerMethodGenerator,
    ModelManagerMethodGenerator,
    ModelPropertyGenerator,
    ModelRelationGenerator,
    PartialUpdateMethodGenerator,
    UpdateMethodGenerator,
)
from .models import (
    ManagerDefinition,
    ManagerMethodDefinition,
    ModelAttributeDefinition,
    ModelDefinition,
    ServiceDefinition,
)
from .shapes import PythonTypeShapeConverter
from .sphinx import ServiceSphinxDocBuilder


class AbstractGenerator:
    def __init__(self, service_generator: "ServiceGenerator") -> None:
        session = botocore.session.get_session()
        #: The :py:class:`ServiceGenerator` we're generating models for.
        self.service_generator = service_generator
        #: The name of the AWS service we're generating models for.
        self.service_name = service_generator.aws_service_name
        #: The botocraft service definition for our service.
        self.service_def = service_generator.service_def
        #: The botocraft interface definition.  We collect things we need to
        #: know globally here.
        self.interface = service_generator.interface
        #: The botocore service model for our service.
        self.service_model = session.get_service_model(self.service_name)
        #: The documentation formatter we will use to format docstrings.
        self.docformatter = DocumentationFormatter()
        self.classes: OrderedDict[str, str] = OrderedDict()
        self.imports: Set[str] = set()

    @property
    def shapes(self) -> List[str]:
        """
        List the names of all the shapes in the service model.

        Returns:
            A list of shape names.

        """
        return self.service_model.shape_names

    def clear(self) -> None:
        """
        Clear the generated classes and imports.
        """
        self.classes = OrderedDict()
        self.imports = set()

    def get_shape(self, name: str) -> botocore.model.Shape:
        """
        Get a :py:class:`botocore.model.Shape` by name from the service model,
        :py:attr:`service_model`.

        Args:
            name: The name of the shape to retrieve.

        Returns:
            The shape object.

        """
        try:
            return self.service_model.shape_for(name)
        except botocore.model.NoShapeFoundError:
            model_name = self.service_def.resolve_model_name(name)
            return self.service_model.shape_for(model_name)

    def get_metadata(self) -> Dict[str, Any]:
        """
        Get the metadata for the botocore service definition.

        Returns:
            The contents of the metadata attribute.

        """
        return self.service_model.metadata

    def generate(self) -> None:
        raise NotImplementedError


class ModelGenerator(AbstractGenerator):
    """
    Generate pydantic model definitions from botocore shapes.
    """

    def __init__(self, service_generator: "ServiceGenerator") -> None:
        super().__init__(service_generator)
        #: The shape converter we will use to convert botocore shapes to python
        #: types, and build the appropriate python classes for ``StructureShape``
        #: objects.
        self.shape_converter = PythonTypeShapeConverter(service_generator, self)

    def get_model_def(self, model_name: str) -> ModelDefinition:
        """
        Return the :py:class:`ModelDefinition` for a model.

        Notes:
            If there is no human defined model definition for the model, we will
            create a default one.

        Args:
            model_name: The name of the model to get the definition for.

        Returns:
            The model definition.

        """
        if model_name in self.service_def.primary_models:
            defn = self.service_def.primary_models[model_name]
            if defn.readonly:
                defn.base_class = "ReadonlyPrimaryBoto3Model"
            else:
                defn.base_class = "PrimaryBoto3Model"
            return defn
        if model_name in self.service_def.secondary_models:
            defn = self.service_def.secondary_models[model_name]
            if defn.readonly:
                defn.base_class = "ReadonlyBoto3Model"
            else:
                defn.base_class = "Boto3Model"
            return defn
        return ModelDefinition(base_class="Boto3Model", name=model_name)

    def add_extra_fields(
        self, model_name: str, model_fields: Dict[str, ModelAttributeDefinition]
    ) -> Dict[str, ModelAttributeDefinition]:
        """
        Extract extra fields from the output shape of a get or list method.
        These are fields that are in the output of the get/list methods but not
        in the service models.  We add them to the models as readonly fields so
        that we can load them from the API responses.

        Args:
            model_name: the name of the model to add fields to
            model_fields: the botocraft model field definitions for the model

        Returns:
            model_fields with the extra fields added

        """
        model_def = self.get_model_def(model_name)
        if not model_def.output_shape:
            return model_fields
        output_shape = self.get_shape(model_def.output_shape)
        if not hasattr(output_shape, "members"):
            return model_fields
        for field_name, field_shape in output_shape.members.items():
            if field_name not in model_fields:
                model_fields[field_name] = ModelAttributeDefinition(
                    readonly=True, botocore_shape=field_shape
                )
        return model_fields

    def mark_readonly_fields(
        self, model_name: str, model_fields: Dict[str, ModelAttributeDefinition]
    ) -> Dict[str, ModelAttributeDefinition]:
        """
        Mark model fields as readonly if they are not in any of the input shapes
        defined for the model.  Such fields are returned by AWS but are not
        settable by the user.

        Args:
            model_name: The name of the model to mark fields for.
            model_fields: The botocraft model field definitions for the model.

        Returns:
            The updated model fields.

        """
        model_def = self.get_model_def(model_name)
        if not model_def.input_shapes:
            return model_fields
        # First include any fields that were manually set to writable
        # in the botocraft model definition
        writable_fields: Set[str] = {
            field_name
            for field_name in model_def.fields
            if model_def.fields[field_name].readonly is False
        }
        # Now add any fields that are in the input shapes
        for input_shape_name in model_def.input_shapes:
            input_shape = self.get_shape(input_shape_name)
            if hasattr(input_shape, "members"):
                writable_fields.update(input_shape.members.keys())
        # Mark any fields that are not in writable_fields as readonly
        for field_name, field_def in model_fields.items():
            if field_name not in writable_fields:
                field_def.readonly = True
        return model_fields

    def fields(self, model_name: str) -> Dict[str, ModelAttributeDefinition]:
        """
        Return the fields for a botocore shape as a dictionary of field names to
        botocraft field definitions.  This incorporates settings from the
        :py:class:`ModelAttributeDefinitions` from the  ``fields`` attribute of
        the associated model definition, if it exists.

        Note:
            This really only makes sense on :py:class:`botocore.model.StructureShape`
            objects, since they are the only ones that have fields (aka "members").

        Side Effects:
            We set :py:attr:`ModelAttributeDefinition.botocore_shape` on each
            field definition, and we set :py:attr:`ModelAttributeDefinition.readonly`
            if the field is not in any of the input shapes for the model.

        Returns:
            A dictionary of field names to field definitions.

        """
        model_def = self.get_model_def(model_name)
        fields: Dict[str, ModelAttributeDefinition] = deepcopy(model_def.fields)
        model_shape = self.get_shape(model_name)
        if hasattr(model_shape, "members"):
            for field, field_shape in model_shape.members.items():
                if field not in fields:
                    fields[field] = ModelAttributeDefinition()
                    if field in model_shape.required_members:
                        fields[field].required = True
                fields[field].botocore_shape = field_shape
        fields = self.mark_readonly_fields(model_name, fields)
        # These are our manually defined extra fields
        if model_def.extra_fields:
            fields.update(model_def.extra_fields)
        # These are the fields that are in the output shape of the get/list
        # methods but not in the service model shape
        fields = self.add_extra_fields(model_name, fields)
        return fields  # noqa: RET504

    def generate(self) -> None:
        """
        Generate all the service models.
        """
        for model_name in self.service_def.primary_models:
            _ = self.generate_model(model_name)

    def extra_fields(self, model_def: ModelDefinition) -> List[str]:
        """
        Build out the manually defined extra fields for a model.

        Extra fields are exclusively defined in the botocore model definition.
        We add them manually to the model definition by inspecting the response
        shape for the create/get/list/update methods and adding any fields that
        aren't already defined in the service model shape.

        Args:
            model_def: The botocraft model definition for this model

        Returns:
            A list of extra fields.

        """
        fields: List[str] = []
        for field_name, field_def in getattr(model_def, "extra_fields", {}).items():
            if field_def.docstring:
                fields.append(self.docformatter.format_attribute(field_def.docstring))
            field = f"    {field_name}: {field_def.python_type}"
            if field_def.readonly:
                field = re.sub(r"Optional\[(.*)\]", r"\1", field)
                _default = f", default={field_def.default}" if field_def.default else ""
                field += f" = Field(frozen=True{_default})"
            elif field_def.default:
                field += f" = {field_def.default}"
            self.imports.update(field_def.imports)
            fields.append(field)
        return fields

    def get_properties(
        self, model_def: ModelDefinition, base_class: str
    ) -> Optional[str]:
        """
        Handle the special properties and methods for primary models.  A primary
        model is a model that has either ``PrimaryBoto3Model`` or
        ``ReadonlyPrimaryBoto3Model`` as its ``base_class``.   Primary models are
        the main models that represent AWS resources, and are those that users can
        create, update, and delete.

        This means:

        * Adding a ``pk`` property that is an alias for the primary key.
        * Maybe adding a ``arn`` property that is an alias for the ARN key.
        * Maybe adding a ``name`` property that is an alias for the name key.
        * Adding any extra properties that were defined in the model definition.
        * Adding any relations to other models that were defined in the model
          definition.
        * Adding any manager shortcut methods that were defined in the model
          definition.

        Args:
            model_def: the botocraft model definition for this model
            base_class: the base class for this model

        Returns:
            The properties for this model, or ``None`` if this is not a primary
            model.

        """
        properties: str = ""
        if base_class in ["PrimaryBoto3Model", "ReadonlyPrimaryBoto3Model"]:
            assert (
                model_def.primary_key or "pk" in model_def.properties
            ), f'Primary service model "{model_def.name}" has no primary key defined'

            if "pk" not in model_def.properties and model_def.primary_key:
                # There is no ``pk`` property, in the ``properties:`` section,
                # but there is a ``primary_key:`` attribute.  We need to add a
                # ``pk`` property that is an alias for the primary key.
                properties = f'''
    @property
    def pk(self) -> Optional[str]:
        """
        Return the primary key of the model.   This is the value of the
        :py:attr:`{model_def.primary_key}` attribute.

        Returns:
            The primary key of the model instance.
        """
        return self.{model_def.primary_key}
'''
            if model_def.arn_key:
                properties += f'''

    @property
    def arn(self) -> Optional[str]:
        """
        Return the ARN of the model.   This is the value of the
        :py:attr:`{model_def.arn_key}` attribute.

        Returns:
            The ARN of the model instance.
        """
        return self.{model_def.arn_key}
'''

            if model_def.name_key:
                properties += f'''

    @property
    def name(self) -> Optional[str]:
        """
        Return the name of the model.   This is the value of the
        :py:attr:`{model_def.name_key}` attribute.

        Returns:
            The name of the model instance.
        """
        return self.{model_def.name_key}
'''
        # Build any regular properties that were defined in the model definition
        for property_name in model_def.properties:
            if not properties:
                properties = ""
            properties += ModelPropertyGenerator(
                self, model_def.name, property_name
            ).code

        # Now build the relations to other models
        for property_name in model_def.relations:
            if not properties:
                properties = ""
            properties += ModelRelationGenerator(
                self, model_def.name, property_name
            ).code

        # Now build the manager shortcut methods
        for method_name in model_def.manager_methods:
            if not properties:
                properties = ""
            properties += ModelManagerMethodGenerator(
                self.service_generator, model_def.name, method_name
            ).code

        return properties

    def field_type(
        self,
        model_name: str,
        field_name: str,
        field_def: Optional[ModelAttributeDefinition] = None,
        model_shape: Optional[botocore.model.Shape] = None,
        field_shape: Optional[botocore.model.Shape] = None,
    ) -> str:
        """
        Return the python type annotation for a field on a model by combining
        the ``model_shape``, ``field_shape`` and ``field_def``.

        Args:
            model_name: The name of the model to get the field from.
            field_name: The name of the field.

        Keyword Args:
            field_def: The botocraft field definition for the field.  If not
                provided, we will look it up in the model definition.
            model_shape: The shape of the model.  If not provided, we will look
                it up in the service model.
            field_shape: The shape of the field.  If not provided, we will look
                it up in the model shape.

        Returns:
            The python type annotation for the field.

        """
        if not model_shape:
            model_shape = self.get_shape(model_name)
        if not hasattr(model_shape, "members"):
            msg = f"Model {model_name} has no fields."
            raise ValueError(msg)
        if not field_shape:
            field_shape = cast(botocore.model.StructureShape, model_shape).members.get(
                field_name
            )
        if not field_def:
            model_def = self.get_model_def(model_name)
            field_def = model_def.fields.get(field_name, ModelAttributeDefinition())
        python_type = field_def.python_type
        if not field_shape and not python_type:
            msg = (
                f"{model_name}.{field_name} has neither a botocore shape nor a "
                "manually defined python_type."
            )
            raise TypeError(msg)
        if not python_type:
            python_type = self.shape_converter.convert(
                cast(botocore.model.StructureShape, field_shape)
            )
        return self.validate_type(model_name, field_name, field_def, python_type)

    def validate_type(
        self,
        model_name: str,
        field_name: str,
        field_def: ModelAttributeDefinition,
        python_type: str,
    ) -> str:
        """
        After we have guessed the python type for a field, we need to validate
        it to make sure it's not going to cause problems for us later.

        Args:
            model_name: the name of the model
            field_name: the name of the field
            field_def: the botocraft field definition for the field
            python_type: the python type we have guessed for the field

        Raises:
            ValueError: we could not determine the type for the field
            TypeError: we have determined that the type for the field is invalid

        Returns:
            The python type annotation for the field on the model.

        """
        if python_type is None:
            msg = f"Could not determine type for field {field_name}."
            raise ValueError(msg)
        name = field_def.rename if field_def.rename else field_name
        if python_type == name or f"[{name}]" in python_type:
            # If our type annotation is for a model with the same name as the field
            # we'll get recursion errors when trying to import the file.  Quoting
            # the type annotation fixes this sometimes.
            python_type = f'"{python_type}"'
        if field_def.readonly and (
            python_type == f'"{name}"' or f'["{name}"]' in python_type
        ):
            # If the field is readonly, and the type is equal to the field name,
            # even if it is quoted, we will get a "TypeError: forward references
            # must evaluate to types".  This happens because when the field is
            # readonly, we set it equal to ``Field(frozen=True, default=None)``.
            # This causes python typing a lot of consternation, and it throws
            # the TypeError.
            msg = (
                f"Field {model_name}.{name} has type equal to its name, "
                'but is marked as readonly.  This will cause a "TypeError: forward '
                'references must evaluate to types". Fix this in '
                f"botocraft/data/{self.service_generator.aws_service_name}/models.yml ."
                "by either giving an alternate_name for the model named {name} or "
                f"by renaming the field {model_name}.{name} with the rename attribute."
            )
            raise TypeError(msg)
        if not field_def.required and (
            python_type == f'"{name}"' or f'["{name}"]' in python_type
        ):
            # If the field is optional with a None default value, and the type
            # is equal to the field name, pydantic will throw an exception when
            # trying to load data into that field: "ValidationError: Input
            # should be None".  This happens even the type is quoted.
            msg = (
                f"Field {model_name}.{name} has type equal to its name, "
                'but is marked as optional.  This will cause a "ValidationError: Input '
                'should be None" exception from pydantic when trying to load data '
                "into this field.  Fix this in "
                f"botocraft/data/{self.service_generator.aws_service_name}/models.yml ."
                "by either giving an alternate_name for the model named {name} or "
                f"by renaming the field {model_name}.{name} with the rename attribute."
            )
            raise TypeError(msg)

        return python_type

    def generate_model(  # noqa: PLR0912, PLR0915
        self, model_name: str, model_shape: Optional[botocore.model.Shape] = None
    ) -> str:
        """
        Generate the code for a single model and its dependent models and save
        them to :py:attr:`classes`.

        Args:
            model_name: The name of the model to generate. This will be the
                name of the class.

        Keyword Args:
            model_shape: The botocore shape to generate the model for.  If not provided,
                we will look it up in the service model.

        Side Effects:
            This may add new models to :py:attr:`classes` and new imports to
            :py:attr:`imports`.

        Returns:
            The name of the model class that was generated.

        """
        # The list of fields for this model
        fields: List[str] = []

        # Get the model definition for this model from the service definition
        # file in ``botocraft/data/<service_name>/models.yml``.
        model_def = self.get_model_def(model_name)
        # Get the base class for this model
        base_class = cast(str, model_def.base_class)
        # This needs to be up here before we check for alternate names
        model_fields = self.fields(model_name)
        if not model_shape:
            model_shape = self.get_shape(model_name)
        if model_def.alternate_name:
            model_name = model_def.alternate_name

        if hasattr(model_shape, "members"):
            # Get the list of all the fields for this model, along with their
            # botocode definitions.  This includes fields from the botocore model,
            # and any extra fields we have defined in the botocraft model
            # definition.
            for field_name, field_def in model_fields.items():
                # The botocore shape for this field.  We determined this in
                # :py:meth:`fields`.
                field_shape = field_def.botocore_shape
                # Whether this field is required
                required: bool = (
                    field_name in model_shape.required_members
                ) or field_def.required
                docstring = field_def.docstring
                # Deterimine the python type for this field
                if field_shape:
                    python_type = self.field_type(
                        model_name,
                        field_name,
                        model_shape=model_shape,
                        field_def=field_def,
                        field_shape=field_shape,
                    )
                    if not docstring:
                        docstring = cast(str, field_shape.documentation)
                else:
                    assert field_def.python_type, (
                        f"Field {field_name} in model {model_name} has no botocore "
                        "shape or python type"
                    )
                    python_type = field_def.python_type
                default = None
                if not required:
                    if not field_def.readonly and not field_def.rename:
                        python_type = f"Optional[{python_type}]"
                    default = "None" if field_def.default is None else field_def.default

                # Determine whether we need to add a pydantic.Field class
                # instance as the value for this field.
                needs_field_class: bool = False
                field_class_args: List[str] = []
                if default:
                    field_class_args.append(f"default={default}" if default else "")
                field_line = f"    {field_name}: {python_type}"
                if field_def.rename:
                    # We need it to be a pydantic Field class instance so that we can
                    # set the serialization_alias attribute.
                    needs_field_class = True
                    field_line = f"    {field_def.rename}: {python_type}"
                    field_class_args.append(f'serialization_alias="{field_name}"')
                if field_def.readonly:
                    # We need it to be a pydantic Field class instance so that we can
                    # set the frozen attribute.
                    needs_field_class = True
                    field_class_args.append("frozen=True")
                if needs_field_class:
                    field_line += f' = Field({", ".join(field_class_args)})'
                elif default:
                    field_line += f" = {field_def.default}"
                fields.append(field_line)
                # Add the docstring for this field
                if docstring:
                    fields.append(self.docformatter.format_attribute(docstring))

            # Add any botocraft defined properties.  This includes relations.
            properties = self.get_properties(model_def, base_class)

            # Add any botocraft defined mixins to the class inheritance
            if model_def.mixins:
                for mixin in model_def.mixins:
                    self.imports.add(f"from {mixin.import_path} import {mixin.name}")
                base_class = ", ".join(
                    [mixin.name for mixin in model_def.mixins] + [base_class]
                )

            # See if we need to add the TagsDictMixin
            tag_class: Optional[str] = None
            has_tags: bool = False
            field_names = {name.lower(): name for name in model_fields}
            if "tags" in field_names:
                # We do have tags defined in the model definition
                has_tags = True
                tag_attr = field_names["tags"]
                if tag_attr == "tags":
                    if model_fields[tag_attr].rename != "Tags":
                        warnings.warn(  # noqa: B028
                            f'Model {model_name} has a field named "tags".  Rename it '
                            'to "Tags" in the model definition.'
                        )
                # Extract the name of the tag class from the python type annotation.
                # Different services use different types for tags.
                tag_class = self.field_type(
                    model_name, tag_attr, model_fields[tag_attr]
                )
                tag_class = re.sub(r"List\[(.*)\]", r"\1", tag_class)
                tag_class = re.sub(r'"(.*)"', r"\1", tag_class)
            if has_tags:
                base_class = f"TagsDictMixin, {base_class}"

            # Actually build the class code
            code: str = f"class {model_name}({base_class}):\n"
            docstring = self.docformatter.format_docstring(model_shape)
            if docstring:
                code += f'    """{docstring}"""\n'
            if has_tags:
                code += f"    tag_class: ClassVar[Type] = {tag_class}\n"
            if "PrimaryBoto3Model" in base_class:
                if model_def.alternate_name:
                    manager_name = f"{model_def.alternate_name}Manager"
                else:
                    manager_name = f"{model_name}Manager"
                code += f"    manager_class: ClassVar[Type[Boto3ModelManager]] = {manager_name}\n\n"  # noqa: E501
            if fields:
                code += "\n".join(fields)
            if properties:
                code += f"\n{properties}"
            if not fields and not properties:
                code += "    pass"
            self.classes[model_name] = code
        return model_name


class ManagerGenerator(AbstractGenerator):
    """
    Generates the code for the manager class for a service.

    Args:
        service_generator: The :py:class:`ServiceGenerator` we're generating
            models for.

    """

    #: A mapping of botocore operation names to the method generator class that
    #: will generate the code for that method.
    METHOD_GENERATORS: Final[Dict[str, Type[ManagerMethodGenerator]]] = {
        "create": CreateMethodGenerator,
        "update": UpdateMethodGenerator,
        "partial_update": PartialUpdateMethodGenerator,
        "delete": DeleteMethodGenerator,
        "get": GetMethodGenerator,
        "get_many": GetManyMethodGenerator,
        "list": ListMethodGenerator,
    }

    def __init__(self, service_generator: "ServiceGenerator") -> None:
        super().__init__(service_generator)
        self.model_generator = self.service_generator.model_generator
        self.shape_converter = self.model_generator.shape_converter
        self.client = boto3.client(self.service_name)  # type: ignore[call-overload]

    def generate_manager(self, model_name: str, manager_def: ManagerDefinition) -> None:
        """
        Generate the code for a single manager, and its dependent response
        classes and save them to :py:attr:`classes`.

        Args:
            model_name: The name of the model to generate the manager for.
            manager_def: The botocraft manager definition for the manager.

        """
        methods: OrderedDict[str, str] = OrderedDict()
        for method_name, method_def in manager_def.methods.items():
            generator = self.get_method_generator(model_name, method_name, method_def)
            methods[method_name] = generator.code
        method_code = "\n\n".join(methods.values())
        base_class = "Boto3ModelManager"
        model_def = self.model_generator.get_model_def(model_name)
        if model_def.alternate_name:
            manager_name = f"{model_def.alternate_name}Manager"
        else:
            manager_name = f"{model_name}Manager"

        # Add any botocraft defined mixins to the class inheritance
        if manager_def.mixins:
            for mixin in manager_def.mixins:
                self.imports.add(f"from {mixin.import_path} import {mixin.name}")
            base_class = ", ".join(
                [mixin.name for mixin in manager_def.mixins] + [base_class]
            )

        # If this is a readonly manager, we need to use the readonly manager
        # base class
        if manager_def.readonly:
            base_class = "ReadonlyBoto3ModelManager"
        code = f"""


class {manager_name}({base_class}):

    service_name: str = '{self.service_name}'

{method_code}
"""
        self.classes[manager_name] = code

    def get_method_generator(
        self, model_name: str, method_name: str, method_def: ManagerMethodDefinition
    ) -> ManagerMethodGenerator:
        """
        Return the appropriate method generator class for a given method
        definition.

        Args:
            model_name: the model name for the manager we're generating
            method_name: the name of the method we're generating
            method_def: the method definition for the method we're generating

        Returns:
            A method generator class.

        """
        try:
            method_generator_class = self.METHOD_GENERATORS[method_name]
        except KeyError:
            # We have no specific method generator for this method, so we
            # will use the general method generator.
            generator: ManagerMethodGenerator = GeneralMethodGenerator(
                self, model_name, method_def, method_name=method_name
            )
        else:
            # We have a specific method generator for this method, so we
            # will use that.
            generator = method_generator_class(self, model_name, method_def)

        return generator

    def generate(self) -> None:
        for model_name, manager_def in self.service_def.managers.items():
            self.generate_manager(model_name, manager_def)
        self.imports.update(self.model_generator.imports)


class ServiceGenerator:
    """
    Generate the code for a single AWS service.

    This means:

        * Managers
        * Service Models
        * Request/Response Models

    Args:
        service_def: The :py:class:`ServiceDefinition` for the service we are
            generating code for.

    """

    service_path: Path = Path(__file__).parent.parent / "services"

    def __init__(self, service_def: ServiceDefinition) -> None:
        #: The service definition
        self.service_def = service_def
        #: The botocraft interface object, where we will collect all our global data
        self.interface = service_def.interface
        #: A set of model imports we need to add to the top of the file
        self.imports: Set[str] = {
            "from datetime import datetime",
            "from typing import ClassVar, Type, Optional, Literal, Dict, List, Any, cast",  # noqa: E501
            "from pydantic import Field",
            "from .abstract import Boto3Model, ReadonlyBoto3Model, PrimaryBoto3Model, "
            "ReadonlyPrimaryBoto3Model, Boto3ModelManager, ReadonlyBoto3ModelManager",
            "from botocraft.mixins.tags import TagsDictMixin",
        }
        #: A dictionary of model names to class code.  This is populated by
        #: service models
        self.model_classes: Dict[str, str] = {}
        #: A dictionary of botocore response classes names to class code. This
        #: is populated when we build the manager classes
        self.response_classes: Dict[str, str] = {}
        #: A dictionary of manager classes names to class code. This is populated
        #: when we build the manager classes
        self.manager_classes: Dict[str, str] = {}
        #: The :py:class:`ModelGenerator` class we will use to generate models
        self.model_generator = ModelGenerator(self)
        #: The :py:class:`ManagerGenerator` class we will use to generate managers
        self.manager_generator = ManagerGenerator(self)
        #: The :py:class:`ManagerGenerator` class we will use to generate managers
        self.sphinx_generator = ServiceSphinxDocBuilder(self)

    @property
    def aws_service_name(self) -> str:
        """
        Return the boto3 service name for this service.
        """
        return self.service_def.name

    @property
    def service_full_name(self) -> str:
        """
        Return what AWS thinks the full name of this services is.

        Returns:
            The full name of the service, as defined in the botocore service.

        """
        return self.model_generator.service_model.metadata["serviceId"]

    @property
    def classes(self) -> Dict[str, str]:
        """
        Return a dictionary of all the classes we have generated.
        """
        return {
            **self.model_classes,
            **self.response_classes,
            **self.manager_classes,
        }

    @property
    def code(self) -> str:
        """
        The code for this service.
        """
        imports = "\n".join(list(self.imports))
        model_classes = "\n\n".join(self.model_classes.values())
        response_classes = "\n\n".join(self.response_classes.values())
        manager_classes = "\n\n".join(self.manager_classes.values())
        return f"""
# This file is automatically generated by botocraft.  Do not edit directly.
# pylint: disable=anomalous-backslash-in-string,unsubscriptable-object,line-too-long,arguments-differ,arguments-renamed,unused-import,redefined-outer-name
# pyright: reportUnusedImport=false
# mypy: disable-error-code="index, override, assignment, union-attr, misc"
{imports}

# ===============
# Managers
# ===============

{manager_classes}


# ==============
# Service Models
# ==============

{model_classes}


# =======================
# Request/Response Models
# =======================

{response_classes}

"""  # noqa: E501

    def generate(self) -> None:
        """
        Generate the code for this service.
        """
        # Generate the service models
        self.model_generator.generate()
        self.model_classes = deepcopy(self.model_generator.classes)
        self.imports.update(self.model_generator.imports)
        self.model_generator.clear()

        # Generate the service managers and request/response models
        self.manager_generator.generate()
        self.response_classes = deepcopy(self.model_generator.classes)
        self.manager_classes = deepcopy(self.manager_generator.classes)
        self.imports.update(self.manager_generator.imports)
        self.model_generator.clear()

        # Write the generated code to the output file
        self.write()
        # Write the sphinx documentation for the service
        self.sphinx_generator.write()

        # Update the interface with the manager models we generated
        for model_name in self.model_classes:
            self.interface.add_model(model_name, self.service_def.name)
        for model_name in self.response_classes:
            self.interface.add_model(model_name, self.service_def.name)
        for model_name in self.manager_classes:
            self.interface.add_model(model_name, self.service_def.name)

    def write(self) -> None:
        """
        Write the generated code to the output file, and format it with black,
        isort, and docformatter.

        Args:
            code: the code to write to the output file.

        """
        code = self.code
        # First format the code with black so we fix most of the formatting
        # issues.
        try:
            formatted_code = black.format_str(code, mode=black.FileMode())
        except (KeyError, black.parsing.InvalidInput):  # pylint: disable=c-extension-no-member
            code = [f"{i:04} " + line for i, line in enumerate(code.split("\n"))]
            print("\n".join(code))
            raise
        # Now sort the imports with isort
        formatted_code = isort.code(formatted_code)
        # Finally format the docstrings with docformatter
        formatted_code = Formatter(FormatterArgs(), None, None, None)._format_code(  # noqa: SLF001
            formatted_code
        )
        output_file = self.service_path / f"{self.service_def.name}.py"
        with output_file.open("w", encoding="utf-8") as fd:
            fd.write(formatted_code)



from typing import TYPE_CHECKING, Callable, List, Literal, Optional, cast

if TYPE_CHECKING:
    from botocraft.services import (
        Instance,
        Reservation,
        Tag,
        TagSpecification,
    )


#: The EC2 resource types.  We need this for specifying the proper tag for
#: resources in EC2.
ResourceType = Literal[
    "capacity-reservation",
    "client-vpn-endpoint",
    "customer-gateway",
    "carrier-gateway",
    "coip-pool",
    "dedicated-host",
    "dhcp-options",
    "egress-only-internet-gateway",
    "elastic-ip",
    "elastic-gpu",
    "export-image-task",
    "export-instance-task",
    "fleet",
    "fpga-image",
    "host-reservation",
    "image",
    "import-image-task",
    "import-snapshot-task",
    "instance",
    "instance-event-window",
    "internet-gateway",
    "ipam",
    "ipam-pool",
    "ipam-scope",
    "ipv4pool-ec2",
    "ipv6pool-ec2",
    "key-pair",
    "launch-template",
    "local-gateway",
    "local-gateway-route-table",
    "local-gateway-virtual-interface",
    "local-gateway-virtual-interface-group",
    "local-gateway-route-table-vpc-association",
    "local-gateway-route-table-virtual-interface-group-association",
    "natgateway",
    "network-acl",
    "network-interface",
    "network-insights-analysis",
    "network-insights-path",
    "network-insights-access-scope",
    "network-insights-access-scope-analysis",
    "placement-group",
    "prefix-list",
    "replace-root-volume-task",
    "reserved-instances",
    "route-table",
    "security-group",
    "security-group-rule",
    "snapshot",
    "spot-fleet-request",
    "spot-instances-request",
    "subnet",
    "subnet-cidr-reservation",
    "traffic-mirror-filter",
    "traffic-mirror-session",
    "traffic-mirror-target",
    "transit-gateway",
    "transit-gateway-attachment",
    "transit-gateway-connect-peer",
    "transit-gateway-multicast-domain",
    "transit-gateway-policy-table",
    "transit-gateway-route-table",
    "transit-gateway-route-table-announcement",
    "volume",
    "vpc",
    "vpc-endpoint",
    "vpc-endpoint-connection",
    "vpc-endpoint-service",
    "vpc-endpoint-service-permission",
    "vpc-peering-connection",
    "vpn-connection",
    "vpn-gateway",
    "vpc-flow-log",
    "capacity-reservation-fleet",
    "traffic-mirror-filter-rule",
    "vpc-endpoint-connection-device-type",
    "verified-access-instance",
    "verified-access-group",
    "verified-access-endpoint",
    "verified-access-policy",
    "verified-access-trust-provider",
    "vpn-connection-device-type",
    "vpc-block-public-access-exclusion",
    "ipam-resource-discovery",
    "ipam-resource-discovery-association",
    "instance-connect-endpoint",
]


# -------------
# Mixin classes
# -------------


class EC2TagsManagerMixin:
    """
    A mixin is used on on :py:class:`botocraft.services.ec2.InstanceManager`
    to convert the odd EC2 tag list to a :py:class:`TagSpecification` object.
    """

    def convert_tags(
        self, tags: Optional[List["Tag"]], resource_type: ResourceType
    ) -> Optional["TagSpecification"]:
        """
        Given a TagList, convert it to a TagSpecification with ResourceType of
        ``resource_type``.

        Args:
            tags: the list of :py:class:`Tag` objects to convert.
            resource_type: the EC2 resource type.

        Returns:
            A :py:class:`TagSpecification` object, or ``None`` if ``tags`` is
            ``None``.

        """
        from botocraft.services import TagSpecification

        if tags is None:
            return None
        return TagSpecification(ResourceType=resource_type, Tags=tags)


class SecurityGroupModelMixin:
    """
    A mixin is used on :py:class:`botocraft.services.ec2.SecurityGroup` to
    enhance the ``.save()`` method to allow for managing ingress and egress
    rules at the same time as saving the security group.

    Normally this is done with several boto3 calls, but this mixin allows for a
    single call to ``.save()`` to create a security group and manage the rules.
    """

    def save(self, **kwargs):
        """
        Save the model.  For security groups, ingress rules are managed via
        separate boto3 calls than the security group itself.  This override of
        the ``save`` method will allow the user to create a security group and
        add ingress rules in one step.
        """
        # TODO: this needs to be enhanced to handle egress rules as well.
        if not self.pk:
            group_id = self.objects.create(self, **kwargs)
            self.objects.authorize_ingress(group_id, self.IpPermissions, **kwargs)
        else:
            old_obj = self.objects.get(self.pk)
            if self.IpPermissions != old_obj.IpPermissions:
                if old_obj.IpPermissions:
                    self.objects.revoke_ingress(
                        self.pk, old_obj.IpPermissions, **kwargs
                    )
                if self.IpPermissions:
                    self.objects.authorize_ingress(
                        self.pk, self.IpPermissions, **kwargs
                    )


# ----------
# Decorators
# ----------


def ec2_instances_only(
    func: Callable[..., List["Reservation"]],
) -> Callable[..., List["Instance"]]:
    """
    Wraps a boto3 method that returns a list of :py:class:`Reservation` objects
    to return a list of :py:class:`Instance` objects instead.
    """

    def wrapper(*args, **kwargs) -> List["Instance"]:
        reservations = func(*args, **kwargs)
        instances: List["Instance"] = []  # noqa: UP037
        for reservation in reservations:
            instances.extend(cast(List["Instance"], reservation.Instances))
        return instances

    return wrapper


def ec2_instance_only(
    func: Callable[..., Optional["Reservation"]],
) -> Callable[..., Optional["Instance"]]:
    """
    Wraps a boto3 method that returns a list of :py:class:`Reservation` objects
    to return a single :py:class:`Instance` object instead.
    """

    def wrapper(*args, **kwargs) -> Optional["Instance"]:
        reservation = func(*args, **kwargs)
        if not reservation:
            return None
        return cast(List["Instance"], reservation.Instances)[0]

    return wrapper



# This file is automatically generated by botocraft.  Do not edit directly.
# pylint: disable=anomalous-backslash-in-string,unsubscriptable-object,line-too-long,arguments-differ,arguments-renamed,unused-import,redefined-outer-name
# pyright: reportUnusedImport=false
# mypy: disable-error-code="index, override, assignment, union-attr, misc"
from collections import OrderedDict
from datetime import datetime
from functools import cached_property
from typing import Any, ClassVar, Dict, List, Literal, Optional, Type, cast

from pydantic import Field

from botocraft.mixins.ecr import (ECRImageMixin, RepositoryMixin,
                                  image_list_images_ecr_images_only,
                                  repo_list_images_ecr_images_only)
from botocraft.mixins.tags import TagsDictMixin
from botocraft.services.common import Tag

from .abstract import (Boto3Model, Boto3ModelManager, PrimaryBoto3Model,
                       ReadonlyBoto3Model, ReadonlyBoto3ModelManager,
                       ReadonlyPrimaryBoto3Model)

# ===============
# Managers
# ===============


class RepositoryManager(Boto3ModelManager):

    service_name: str = "ecr"

    def create(
        self, model: "Repository", tags: Optional[List[Tag]] = None
    ) -> "Repository":
        """
        Create an ECR repository.

        Args:
            model: The :py:class:``Repository`` to create.

        Keyword Args:
            tags: The metadata that you apply to the repository to help you categorize
                and organize them. Each tag consists of a key and an optional value, both
                of which you define. Tag keys can have a maximum character length of 128
                characters, and tag values can have a maximum length of 256 characters.
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            repositoryName=data.get("repositoryName"),
            registryId=data.get("registryId"),
            tags=self.serialize(tags),
            imageTagMutability=data.get("imageTagMutability"),
            imageScanningConfiguration=data.get("imageScanningConfiguration"),
            encryptionConfiguration=data.get("encryptionConfiguration"),
        )
        _response = self.client.create_repository(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = CreateRepositoryResponse(**_response)

        self.sessionize(response.repository)
        return cast("Repository", response.repository)

    def delete(
        self,
        repositoryName: str,
        *,
        registryId: Optional[str] = None,
        force: Optional[bool] = None
    ) -> "Repository":
        """
        Delete an ECR repository.

        Args:
            repositoryName: The name of the repository to delete.

        Keyword Args:
            registryId: The Amazon Web Services account ID associated with the registry
                that contains the repository to delete. If you do not specify a registry,
                the default registry is assumed.
            force: If true, deleting the repository force deletes the contents of the
                repository. If false, the repository must be empty before attempting to
                delete it.
        """
        args: Dict[str, Any] = dict(
            repositoryName=self.serialize(repositoryName),
            registryId=self.serialize(registryId),
            force=self.serialize(force),
        )
        _response = self.client.delete_repository(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = DeleteRepositoryResponse(**_response)
        return cast(Repository, response.repository)

    def get(
        self, repositoryName: str, *, registryId: Optional[str] = None
    ) -> Optional["Repository"]:
        """
        Describes image repositories in a registry.

        Args:
            repositoryName: The name of the ECR repository to describe.

        Keyword Args:
            registryId: The Amazon Web Services account ID associated with the registry
                that contains the repositories to be described. If you do not specify a
                registry, the default registry is assumed.
        """
        args: Dict[str, Any] = dict(
            registryId=self.serialize(registryId),
            repositoryNames=self.serialize([repositoryName]),
        )
        _response = self.client.describe_repositories(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = DescribeRepositoriesResponse(**_response)

        if response and response.repositories:
            self.sessionize(response.repositories[0])
            return response.repositories[0]
        return None

    def list(
        self,
        *,
        registryId: Optional[str] = None,
        repositoryNames: Optional[List[str]] = None
    ) -> List["Repository"]:
        """
        Describes image repositories in a registry.

        Keyword Args:
            registryId: The Amazon Web Services account ID associated with the registry
                that contains the repositories to be described. If you do not specify a
                registry, the default registry is assumed.
            repositoryNames: A list of repositories to describe. If this parameter is
                omitted, then all repositories in a registry are described.
        """
        paginator = self.client.get_paginator("describe_repositories")
        args: Dict[str, Any] = dict(
            registryId=self.serialize(registryId),
            repositoryNames=self.serialize(repositoryNames),
        )
        response_iterator = paginator.paginate(
            **{k: v for k, v in args.items() if v is not None}
        )
        results: List["Repository"] = []
        for _response in response_iterator:
            response = DescribeRepositoriesResponse(**_response)
            if response.repositories:
                results.extend(response.repositories)
            else:
                break
        self.sessionize(results)
        return results

    @repo_list_images_ecr_images_only
    def list_images(
        self,
        repositoryName: str,
        *,
        registryId: Optional[str] = None,
        filter: Optional["ListImagesFilter"] = None
    ) -> List["ImageIdentifier"]:
        """
        Lists all the image IDs for the specified repository.

        Args:
            repositoryName: The repository with image IDs to be listed.

        Keyword Args:
            registryId: The Amazon Web Services account ID associated with the registry
                that contains the repository in which to list images. If you do not specify
                a registry, the default registry is assumed.
            filter: The filter key and value with which to filter your ``ListImages``
                results.
        """
        paginator = self.client.get_paginator("list_images")
        args: Dict[str, Any] = dict(
            repositoryName=self.serialize(repositoryName),
            registryId=self.serialize(registryId),
            filter=self.serialize(filter),
        )
        response_iterator = paginator.paginate(
            **{k: v for k, v in args.items() if v is not None}
        )

        results: List["ImageIdentifier"] = []

        for _response in response_iterator:
            response = ListImagesResponse(**_response)
            if response.imageIds is not None:
                results.extend(response.imageIds)
            else:
                break

        self.sessionize(results)
        return cast(List["ImageIdentifier"], results)

    def get_images(
        self,
        repositoryName: str,
        imageIds: List["ImageIdentifier"],
        *,
        acceptedMediaTypes: List[str] = [
            "application/vnd.docker.distribution.manifest.v2+json"
        ]
    ) -> Optional[List["ECRImage"]]:
        """
        Use this method when you want to get just a few images from the
        repository. If you want to get all images, use the ````list_images````
        method.

        Args:
            repositoryName: The repository that contains the images to describe.
            imageIds: A list of image ID references that correspond to images to
                describe. The format of the ``imageIds`` reference is ``imageTag=tag`` or
                ``imageDigest=digest``.

        Keyword Args:
            acceptedMediaTypes: The accepted media types for the request.
        """
        args: Dict[str, Any] = dict(
            repositoryName=self.serialize(repositoryName),
            imageIds=self.serialize(imageIds),
            acceptedMediaTypes=self.serialize(acceptedMediaTypes),
        )
        _response = self.client.batch_get_image(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = BatchGetImageResponse(**_response)

        results: Optional[List["ECRImage"]] = None
        if response is not None:
            results = response.images

        self.sessionize(results)
        return cast(Optional[List["ECRImage"]], results)

    def get_image(
        self,
        repositoryName: str,
        imageId: "ImageIdentifier",
        *,
        acceptedMediaTypes: List[str] = [
            "application/vnd.docker.distribution.manifest.v2+json"
        ]
    ) -> "ECRImage":
        """
        Gets detailed information for an image. Images are specified with
        either an ``imageTag`` or ``imageDigest``.

        Args:
            repositoryName: The repository that contains the images to describe.
            imageId: The image ID or tag to describe. The format of the imageId
                reference is ````imageTag=tag```` or ````imageDigest=digest````

        Keyword Args:
            acceptedMediaTypes: The accepted media types for the request.
        """
        args: Dict[str, Any] = dict(
            repositoryName=self.serialize(repositoryName),
            imageIds=self.serialize([imageId]),
            acceptedMediaTypes=self.serialize(acceptedMediaTypes),
        )
        _response = self.client.batch_get_image(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = BatchGetImageResponse(**_response)

        results: "ECRImage" = None
        if response is not None:
            results = response.images[0]

        self.sessionize(results)
        return cast("ECRImage", results)


class ECRImageManager(Boto3ModelManager):

    service_name: str = "ecr"

    def get(
        self,
        repositoryName: str,
        imageId: "ImageIdentifier",
        *,
        acceptedMediaTypes: List[str] = [
            "application/vnd.docker.distribution.manifest.v2+json"
        ]
    ) -> Optional["ECRImage"]:
        """
        Gets detailed information for an image. Images are specified with
        either an ``imageTag`` or ``imageDigest``.

        Args:
            repositoryName: The repository that contains the images to describe.
            imageId: The image ID or tag to describe. The format of the imageId
                reference is ````imageTag=tag```` or ````imageDigest=digest````

        Keyword Args:
            acceptedMediaTypes: The accepted media types for the request.
        """
        args: Dict[str, Any] = dict(
            repositoryName=self.serialize(repositoryName),
            imageIds=self.serialize([imageId]),
            acceptedMediaTypes=self.serialize(acceptedMediaTypes),
        )
        _response = self.client.batch_get_image(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = BatchGetImageResponse(**_response)

        if response and response.images:
            self.sessionize(response.images[0])
            return response.images[0]
        return None

    def get_many(
        self,
        repositoryName: str,
        imageIds: List["ImageIdentifier"],
        *,
        acceptedMediaTypes: List[str] = [
            "application/vnd.docker.distribution.manifest.v2+json"
        ]
    ) -> "BatchGetImageResponse":
        """
        Gets detailed information for an image. Images are specified with
        either an ``imageTag`` or ``imageDigest``.

        Args:
            repositoryName: The repository that contains the images to describe.
            imageIds: A list of image ID references that correspond to images to
                describe. The format of the ``imageIds`` reference is ``imageTag=tag`` or
                ``imageDigest=digest``.

        Keyword Args:
            acceptedMediaTypes: The accepted media types for the request.
        """
        args: Dict[str, Any] = dict(
            repositoryName=self.serialize(repositoryName),
            imageIds=self.serialize(imageIds),
            acceptedMediaTypes=self.serialize(acceptedMediaTypes),
        )
        _response = self.client.batch_get_image(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = BatchGetImageResponse(**_response)

        self.sessionize(response)
        return response

    @image_list_images_ecr_images_only
    def list(
        self, repositoryName: str, *, filter: Optional["ListImagesFilter"] = None
    ) -> List["ImageIdentifier"]:
        """
        Lists all the image IDs for the specified repository.

        Args:
            repositoryName: The repository with image IDs to be listed.

        Keyword Args:
            filter: The filter key and value with which to filter your ``ListImages``
                results.
        """
        paginator = self.client.get_paginator("list_images")
        args: Dict[str, Any] = dict(
            repositoryName=self.serialize(repositoryName), filter=self.serialize(filter)
        )
        response_iterator = paginator.paginate(
            **{k: v for k, v in args.items() if v is not None}
        )
        results: List["ImageIdentifier"] = []
        for _response in response_iterator:
            response = ListImagesResponse(**_response)
            if response.imageIds:
                results.extend(response.imageIds)
            else:
                break
        self.sessionize(results)
        return results

    def delete(
        self, repositoryName: str, imageId: "ImageIdentifier"
    ) -> "BatchDeleteImageResponse":
        """
        Deletes a list of specified images within a repository. Images are
        specified with either an ``imageTag`` or ``imageDigest``.

        Args:
            repositoryName: The repository that contains the image to delete.
            imageId: The image ID or tag to delete. The format of the imageId reference
                is ````imageTag=tag```` or ````imageDigest=digest````
        """
        args: Dict[str, Any] = dict(
            repositoryName=self.serialize(repositoryName),
            imageIds=self.serialize([imageId]),
        )
        _response = self.client.batch_delete_image(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = BatchDeleteImageResponse(**_response)
        return response

    def replication_status(
        self, repositoryName: str, imageId: "ImageIdentifier"
    ) -> "DescribeImageReplicationStatusResponse":
        """
        Returns the replication status for a specified image.

        Args:
            repositoryName: The name of the repository that the image is in.
            imageId: An object with identifying information for an image in an Amazon
                ECR repository.
        """
        args: Dict[str, Any] = dict(
            repositoryName=self.serialize(repositoryName),
            imageId=self.serialize(imageId),
        )
        _response = self.client.describe_image_replication_status(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = DescribeImageReplicationStatusResponse(**_response)

        results: "DescribeImageReplicationStatusResponse" = None
        if response is not None:
            results = response

        self.sessionize(results)
        return cast("DescribeImageReplicationStatusResponse", results)

    def scan_findings(
        self, repositoryName: str, imageId: "ImageIdentifier"
    ) -> List["DescribeImageScanFindingsResponse"]:
        """
        Returns the scan findings for the specified image.

        Args:
            repositoryName: The repository for the image for which to describe the scan
                findings.
            imageId: An object with identifying information for an image in an Amazon
                ECR repository.
        """
        paginator = self.client.get_paginator("describe_image_scan_findings")
        args: Dict[str, Any] = dict(
            repositoryName=self.serialize(repositoryName),
            imageId=self.serialize(imageId),
        )
        response_iterator = paginator.paginate(
            **{k: v for k, v in args.items() if v is not None}
        )

        results: List["DescribeImageScanFindingsResponse"] = []

        for _response in response_iterator:
            response = DescribeImageScanFindingsResponse(**_response)
            results.append(response)

        self.sessionize(results)
        return cast(List["DescribeImageScanFindingsResponse"], results)


# ==============
# Service Models
# ==============


class ImageScanningConfiguration(Boto3Model):
    """
    The image scanning configuration for a repository.
    """

    scanOnPush: Optional[bool] = None
    """
    The setting that determines whether images are scanned after being pushed
    to a repository.

    If set to ``true``, images will be scanned after being pushed. If
    this parameter is not specified, it will default to ``false`` and images will
    not be scanned unless a scan is manually started with the [API_StartImageScan]
    (https://docs.aws.amazon.com/AmazonECR/latest/APIReference/API_StartImageScan.h
    tml) API.
    """


class EncryptionConfiguration(Boto3Model):
    """
    The encryption configuration for the repository.

    This determines how the contents of your repository are encrypted at rest.
    """

    encryptionType: Literal["AES256", "KMS", "KMS_DSSE"]
    """
    The encryption type to use.
    """
    kmsKey: Optional[str] = None
    """
    If you use the ``KMS`` encryption type, specify the KMS key to use for
    encryption.

    The alias, key ID, or full ARN of the KMS key can be specified. The key
    must exist in the same Region as the repository. If no key is specified,
    the default Amazon Web Services managed KMS key for Amazon ECR will be
    used.
    """


class Repository(RepositoryMixin, PrimaryBoto3Model):
    """
    An object representing a repository.
    """

    manager_class: ClassVar[Type[Boto3ModelManager]] = RepositoryManager

    repositoryName: str
    """
    The name of the repository.
    """
    imageTagMutability: Literal["MUTABLE", "IMMUTABLE"]
    """
    The tag mutability setting for the repository.
    """
    imageScanningConfiguration: ImageScanningConfiguration
    """
    The image scanning configuration for a repository.
    """
    repositoryArn: str = Field(default=None, frozen=True)
    """
    The Amazon Resource Name (ARN) that identifies the repository.

    The ARN contains
    the ``arn:aws:ecr`` namespace, followed by the region of the repository, Amazon
    Web Services account ID of the repository owner, repository namespace, and
    repository name. For example, ``arn:aws:ecr:region:012345678910:repository-
    namespace/repository-name``.
    """
    registryId: Optional[str] = None
    """
    The Amazon Web Services account ID associated with the registry that
    contains the repository.
    """
    repositoryUri: str = Field(default=None, frozen=True)
    """
    The URI for the repository.

    You can use this URI for container image ``push``
    and ``pull`` operations.
    """
    createdAt: datetime = Field(default=None, frozen=True)
    """
    The date and time, in JavaScript date format, when the repository was
    created.
    """
    encryptionConfiguration: Optional[EncryptionConfiguration] = None
    """
    The encryption configuration for the repository.

    This determines how the contents of your repository are encrypted at rest.
    """

    @property
    def pk(self) -> Optional[str]:
        """
        Return the primary key of the model.   This is the value of the
        :py:attr:`repositoryName` attribute.

        Returns:
            The primary key of the model instance.
        """
        return self.repositoryName

    @property
    def arn(self) -> Optional[str]:
        """
        Return the ARN of the model.   This is the value of the
        :py:attr:`repositoryArn` attribute.

        Returns:
            The ARN of the model instance.
        """
        return self.repositoryArn

    @property
    def name(self) -> Optional[str]:
        """
        Return the name of the model.   This is the value of the
        :py:attr:`repositoryName` attribute.

        Returns:
            The name of the model instance.
        """
        return self.repositoryName


class ImageIdentifier(Boto3Model):
    """
    An object containing the image tag and image digest associated with an
    image.
    """

    imageDigest: Optional[str] = None
    """
    The ``sha256`` digest of the image manifest.
    """
    imageTag: Optional[str] = None
    """
    The tag used for the image.
    """


class ECRImage(ECRImageMixin, ReadonlyPrimaryBoto3Model):
    """
    An object representing an Amazon ECR image.
    """

    manager_class: ClassVar[Type[Boto3ModelManager]] = ECRImageManager

    registryId: Optional[str] = None
    """
    The Amazon Web Services account ID associated with the registry containing
    the image.
    """
    repositoryName: Optional[str] = None
    """
    The name of the repository associated with the image.
    """
    imageId: ImageIdentifier = Field(default=None, frozen=True)
    """
    An object containing the image tag and image digest associated with an
    image.
    """
    imageManifest: str = Field(default=None, frozen=True)
    """
    The image manifest associated with the image.
    """
    imageManifestMediaType: str = Field(default=None, frozen=True)
    """
    The manifest media type of the image.
    """

    @property
    def pk(self) -> OrderedDict[str, Any]:

        return OrderedDict(
            {
                "repositoryName": self.repositoryName,
                "imageId": self.imageId,
            }
        )

    @cached_property
    def repository(self) -> Optional["Repository"]:
        """
        Return the :py:class:`Repository` object that this image belongs to.

        .. note::

            The output of this property is cached on the model instance, so
            calling this multiple times will not result in multiple calls to the
            AWS API.   If you need a fresh copy of the data, you can re-get the
            model instance from the manager.
        """

        try:
            pk = OrderedDict(
                {
                    "repositoryName": self.repositoryName,
                }
            )
        except AttributeError:
            return None
        return Repository.objects.using(self.session).get(**pk)  # type: ignore[arg-type]

    def replication_status(self) -> "DescribeImageReplicationStatusResponse":
        """
        Return the replication status for the image.
        """

        return (
            cast(ECRImageManager, self.objects)
            .using(self.session)
            .replication_status(cast(str, self.repositoryName), imageId=self.imageId)
        )

    def scan_findings(self) -> List["DescribeImageScanFindingsResponse"]:
        """
        Return the scan results for the image.
        """

        return (
            cast(ECRImageManager, self.objects)
            .using(self.session)
            .scan_findings(cast(str, self.repositoryName), imageId=self.imageId)
        )


# =======================
# Request/Response Models
# =======================


class CreateRepositoryResponse(Boto3Model):
    repository: Optional[Repository] = None
    """
    The repository that was created.
    """


class DeleteRepositoryResponse(Boto3Model):
    repository: Optional[Repository] = None
    """
    The repository that was deleted.
    """


class DescribeRepositoriesResponse(Boto3Model):
    repositories: Optional[List["Repository"]] = None
    """
    A list of repository objects corresponding to valid repositories.
    """
    nextToken: Optional[str] = None
    """
    The ``nextToken`` value to include in a future ``DescribeRepositories``
    request.

    When the results of a ``DescribeRepositories`` request exceed
    ``maxResults``, this value can be used to retrieve the next page of results.
    This value is ``null`` when there are no more results to return.
    """


class ListImagesFilter(Boto3Model):
    """
    The filter key and value with which to filter your ``ListImages`` results.
    """

    tagStatus: Optional[Literal["TAGGED", "UNTAGGED", "ANY"]] = None
    """
    The tag status with which to filter your ListImages results.

    You can filter
    results based on whether they are ``TAGGED`` or ``UNTAGGED``.
    """


class ListImagesResponse(Boto3Model):
    imageIds: Optional[List["ImageIdentifier"]] = None
    """
    The list of image IDs for the requested repository.
    """
    nextToken: Optional[str] = None
    """
    The ``nextToken`` value to include in a future ``ListImages`` request.

    When the
    results of a ``ListImages`` request exceed ``maxResults``, this value can be
    used to retrieve the next page of results. This value is ``null`` when there
    are no more results to return.
    """


class ImageFailure(Boto3Model):
    """
    An object representing an Amazon ECR image failure.
    """

    imageId: Optional[ImageIdentifier] = None
    """
    The image ID associated with the failure.
    """
    failureCode: Optional[
        Literal[
            "InvalidImageDigest",
            "InvalidImageTag",
            "ImageTagDoesNotMatchDigest",
            "ImageNotFound",
            "MissingDigestAndTag",
            "ImageReferencedByManifestList",
            "KmsError",
            "UpstreamAccessDenied",
            "UpstreamTooManyRequests",
            "UpstreamUnavailable",
        ]
    ] = None
    """
    The code associated with the failure.
    """
    failureReason: Optional[str] = None
    """
    The reason for the failure.
    """


class BatchGetImageResponse(Boto3Model):
    images: Optional[List["ECRImage"]] = None
    """
    A list of image objects corresponding to the image references in the
    request.
    """
    failures: Optional[List["ImageFailure"]] = None
    """
    Any failures associated with the call.
    """


class BatchDeleteImageResponse(Boto3Model):
    imageIds: Optional[List["ImageIdentifier"]] = None
    """
    The image IDs of the deleted images.
    """
    failures: Optional[List["ImageFailure"]] = None
    """
    Any failures associated with the call.
    """


class ImageReplicationStatus(Boto3Model):
    """
    The status of the replication process for an image.
    """

    region: Optional[str] = None
    """
    The destination Region for the image replication.
    """
    registryId: Optional[str] = None
    """
    The Amazon Web Services account ID associated with the registry to which
    the image belongs.
    """
    status: Optional[Literal["IN_PROGRESS", "COMPLETE", "FAILED"]] = None
    """
    The image replication status.
    """
    failureCode: Optional[str] = None
    """
    The failure code for a replication that has failed.
    """


class DescribeImageReplicationStatusResponse(Boto3Model):
    repositoryName: Optional[str] = None
    """
    The repository name associated with the request.
    """
    imageId: Optional[ImageIdentifier] = None
    """
    An object with identifying information for an image in an Amazon ECR
    repository.
    """
    replicationStatuses: Optional[List["ImageReplicationStatus"]] = None
    """
    The replication status details for the images in the specified repository.
    """


class ImageScanStatus(Boto3Model):
    """
    The current state of the scan.
    """

    status: Optional[
        Literal[
            "IN_PROGRESS",
            "COMPLETE",
            "FAILED",
            "UNSUPPORTED_IMAGE",
            "ACTIVE",
            "PENDING",
            "SCAN_ELIGIBILITY_EXPIRED",
            "FINDINGS_UNAVAILABLE",
        ]
    ] = None
    """
    The current state of an image scan.
    """
    description: Optional[str] = None
    """
    The description of the image scan status.
    """


class ECRAttribute(Boto3Model):
    """
    This data type is used in the ImageScanFinding data type.
    """

    key: str
    """
    The attribute key.
    """
    value: Optional[str] = None
    """
    The value assigned to the attribute key.
    """


class ImageScanFinding(Boto3Model):
    """
    Contains information about an image scan finding.
    """

    name: Optional[str] = None
    """
    The name associated with the finding, usually a CVE number.
    """
    description: Optional[str] = None
    """
    The description of the finding.
    """
    uri: Optional[str] = None
    """
    A link containing additional details about the security vulnerability.
    """
    severity: Optional[
        Literal["INFORMATIONAL", "LOW", "MEDIUM", "HIGH", "CRITICAL", "UNDEFINED"]
    ] = None
    """
    The finding severity.
    """
    attributes: Optional[List["ECRAttribute"]] = None
    """
    A collection of attributes of the host from which the finding is generated.
    """


class CvssScore(Boto3Model):
    """
    The CVSS score for a finding.
    """

    baseScore: Optional[float] = None
    """
    The base CVSS score used for the finding.
    """
    scoringVector: Optional[str] = None
    """
    The vector string of the CVSS score.
    """
    source: Optional[str] = None
    """
    The source of the CVSS score.
    """
    version: Optional[str] = None
    """
    The version of CVSS used for the score.
    """


class VulnerablePackage(Boto3Model):
    """
    Information on the vulnerable package identified by a finding.
    """

    arch: Optional[str] = None
    """
    The architecture of the vulnerable package.
    """
    epoch: Optional[int] = None
    """
    The epoch of the vulnerable package.
    """
    filePath: Optional[str] = None
    """
    The file path of the vulnerable package.
    """
    name: Optional[str] = None
    """
    The name of the vulnerable package.
    """
    packageManager: Optional[str] = None
    """
    The package manager of the vulnerable package.
    """
    release: Optional[str] = None
    """
    The release of the vulnerable package.
    """
    sourceLayerHash: Optional[str] = None
    """
    The source layer hash of the vulnerable package.
    """
    version: Optional[str] = None
    """
    The version of the vulnerable package.
    """
    fixedInVersion: Optional[str] = None
    """
    The version of the package that contains the vulnerability fix.
    """


class PackageVulnerabilityDetails(Boto3Model):
    """
    An object that contains the details of a package vulnerability finding.
    """

    cvss: Optional[List["CvssScore"]] = None
    """
    An object that contains details about the CVSS score of a finding.
    """
    referenceUrls: Optional[List[str]] = None
    """
    One or more URLs that contain details about this vulnerability type.
    """
    relatedVulnerabilities: Optional[List[str]] = None
    """
    One or more vulnerabilities related to the one identified in this finding.
    """
    source: Optional[str] = None
    """
    The source of the vulnerability information.
    """
    sourceUrl: Optional[str] = None
    """
    A URL to the source of the vulnerability information.
    """
    vendorCreatedAt: Optional[datetime] = None
    """
    The date and time that this vulnerability was first added to the vendor's
    database.
    """
    vendorSeverity: Optional[str] = None
    """
    The severity the vendor has given to this vulnerability type.
    """
    vendorUpdatedAt: Optional[datetime] = None
    """
    The date and time the vendor last updated this vulnerability in their
    database.
    """
    vulnerabilityId: Optional[str] = None
    """
    The ID given to this vulnerability.
    """
    vulnerablePackages: Optional[List["VulnerablePackage"]] = None
    """
    The packages impacted by this vulnerability.
    """


class Recommendation(Boto3Model):
    """
    An object that contains information about the recommended course of action
    to remediate the finding.
    """

    url: Optional[str] = None
    """
    The URL address to the CVE remediation recommendations.
    """
    text: Optional[str] = None
    """
    The recommended course of action to remediate the finding.
    """


class Remediation(Boto3Model):
    """
    An object that contains the details about how to remediate a finding.
    """

    recommendation: Optional[Recommendation] = None
    """
    An object that contains information about the recommended course of action
    to remediate the finding.
    """


class AwsEcrContainerImageDetails(Boto3Model):
    """
    An object that contains details about the Amazon ECR container image
    involved in the finding.
    """

    architecture: Optional[str] = None
    """
    The architecture of the Amazon ECR container image.
    """
    author: Optional[str] = None
    """
    The image author of the Amazon ECR container image.
    """
    imageHash: Optional[str] = None
    """
    The image hash of the Amazon ECR container image.
    """
    imageTags: Optional[List[str]] = None
    """
    The image tags attached to the Amazon ECR container image.
    """
    platform: Optional[str] = None
    """
    The platform of the Amazon ECR container image.
    """
    pushedAt: Optional[datetime] = None
    """
    The date and time the Amazon ECR container image was pushed.
    """
    registry: Optional[str] = None
    """
    The registry the Amazon ECR container image belongs to.
    """
    repositoryName: Optional[str] = None
    """
    The name of the repository the Amazon ECR container image resides in.
    """


class ResourceDetails(Boto3Model):
    """
    An object that contains details about the resource involved in a finding.
    """

    awsEcrContainerImage: Optional[AwsEcrContainerImageDetails] = None
    """
    An object that contains details about the Amazon ECR container image
    involved in the finding.
    """


class Resource(TagsDictMixin, Boto3Model):
    """
    Details about the resource involved in a finding.
    """

    tag_class: ClassVar[Type] = Dict[str, str]
    Tags: Dict[str, str] = Field(default=None, serialization_alias="tags")
    """
    The tags attached to the resource.
    """
    details: Optional[ResourceDetails] = None
    """
    An object that contains details about the resource involved in a finding.
    """
    id: Optional[str] = None
    """
    The ID of the resource.
    """
    type: Optional[str] = None
    """
    The type of resource.
    """


class CvssScoreAdjustment(Boto3Model):
    """
    Details on adjustments Amazon Inspector made to the CVSS score for a
    finding.
    """

    metric: Optional[str] = None
    """
    The metric used to adjust the CVSS score.
    """
    reason: Optional[str] = None
    """
    The reason the CVSS score has been adjustment.
    """


class CvssScoreDetails(Boto3Model):
    """
    An object that contains details about the CVSS score given to a finding.
    """

    adjustments: Optional[List["CvssScoreAdjustment"]] = None
    """
    An object that contains details about adjustment Amazon Inspector made to
    the CVSS score.
    """
    score: Optional[float] = None
    """
    The CVSS score.
    """
    scoreSource: Optional[str] = None
    """
    The source for the CVSS score.
    """
    scoringVector: Optional[str] = None
    """
    The vector for the CVSS score.
    """
    version: Optional[str] = None
    """
    The CVSS version used in scoring.
    """


class ScoreDetails(Boto3Model):
    """
    An object that contains details of the Amazon Inspector score.
    """

    cvss: Optional[CvssScoreDetails] = None
    """
    An object that contains details about the CVSS score given to a finding.
    """


class EnhancedImageScanFinding(Boto3Model):
    """
    The details of an enhanced image scan.

    This is returned when enhanced scanning is enabled for your private
    registry.
    """

    awsAccountId: Optional[str] = None
    """
    The Amazon Web Services account ID associated with the image.
    """
    description: Optional[str] = None
    """
    The description of the finding.
    """
    findingArn: Optional[str] = None
    """
    The Amazon Resource Number (ARN) of the finding.
    """
    firstObservedAt: Optional[datetime] = None
    """
    The date and time that the finding was first observed.
    """
    lastObservedAt: Optional[datetime] = None
    """
    The date and time that the finding was last observed.
    """
    packageVulnerabilityDetails: Optional[PackageVulnerabilityDetails] = None
    """
    An object that contains the details of a package vulnerability finding.
    """
    remediation: Optional[Remediation] = None
    """
    An object that contains the details about how to remediate a finding.
    """
    resources: Optional[List["Resource"]] = None
    """
    Contains information on the resources involved in a finding.
    """
    score: Optional[float] = None
    """
    The Amazon Inspector score given to the finding.
    """
    scoreDetails: Optional[ScoreDetails] = None
    """
    An object that contains details of the Amazon Inspector score.
    """
    severity: Optional[str] = None
    """
    The severity of the finding.
    """
    status: Optional[str] = None
    """
    The status of the finding.
    """
    title: Optional[str] = None
    """
    The title of the finding.
    """
    type: Optional[str] = None
    """
    The type of the finding.
    """
    updatedAt: Optional[datetime] = None
    """
    The date and time the finding was last updated at.
    """
    fixAvailable: Optional[str] = None
    """
    Details on whether a fix is available through a version update.

    This value can
    be ``YES``, ``NO``, or ``PARTIAL``. A ``PARTIAL`` fix means that some, but not
    all, of the packages identified in the finding have fixes available through
    updated versions.
    """
    exploitAvailable: Optional[str] = None
    """
    If a finding discovered in your environment has an exploit available.
    """


class ImageScanFindings(ReadonlyBoto3Model):
    """
    The information contained in the image scan findings.
    """

    findingSeverityCounts: Optional[
        Optional[
            Dict[
                Literal[
                    "INFORMATIONAL",
                    "LOW",
                    "MEDIUM",
                    "HIGH",
                    "CRITICAL",
                    "UNDEFINED",
                    "UNTRIAGED",
                ],
                int,
            ]
        ]
    ] = None
    """
    The image vulnerability counts, sorted by severity.
    """
    imageScanCompletedAt: Optional[datetime] = None
    """
    The time of the last completed image scan.
    """
    vulnerabilitySourceUpdatedAt: Optional[datetime] = None
    """
    The time when the vulnerability data was last scanned.
    """
    findings: Optional[List["ImageScanFinding"]] = None
    """
    The findings from the image scan.
    """
    enhancedFindings: Optional[List["EnhancedImageScanFinding"]] = None
    """
    Details about the enhanced scan findings from Amazon Inspector.
    """


class DescribeImageScanFindingsResponse(Boto3Model):
    registryId: Optional[str] = None
    """
    The registry ID associated with the request.
    """
    repositoryName: Optional[str] = None
    """
    The repository name associated with the request.
    """
    imageId: Optional[ImageIdentifier] = None
    """
    An object with identifying information for an image in an Amazon ECR
    repository.
    """
    imageScanStatus: Optional[ImageScanStatus] = None
    """
    The current state of the scan.
    """
    imageScanFindings: Optional[ImageScanFindings] = None
    """
    The information contained in the image scan findings.
    """
    nextToken: Optional[str] = None
    """
    The ``nextToken`` value to include in a future
    ``DescribeImageScanFindings`` request.

    When the results of a ``DescribeImageScanFindings`` request exceed
    ``maxResults``, this value can be used to retrieve the next page of results.
    This value is null when there are no more results to return.
    """



from .base import ManagerMethodGenerator


class GetManyMethodGenerator(ManagerMethodGenerator):
    """
    Generate the code for the ``get_many`` method.  This differs from the
    ``list`` method in that sometimes the boto3 operation that handles listing
    does not actually return all the data we need to create our model.  In those
    cases, there is usually a separate boto3 operation that returns lists of
    objects that we can use to get the full data we need.

    When this happens, the ``get_many`` boto3 operation typically cannot be
    paginated, otherwise we would have used it as the ``list`` method.
    """

    method_name: str = "get_many"

    @property
    def return_type(self) -> str:
        """
        For get_many methods, we return a list of model instances, not the response
        model, unless it's overriden in our botocraft method config, in which
        case we return that.

        Thus we need to change the return type to a list of the model.

        Returns:
            The name of the return type class.

        """
        _ = self.response_class
        return_type = f'List["{self.model_name}"]'
        if self.method_def.return_type:
            return_type = self.method_def.return_type
        return return_type

    @property
    def body(self) -> str:
        code = f"""
        {self.operation_args}
        {self.operation_call}
"""
        if self.response_attr is not None:
            code += f"""
        self.sessionize(respsonse.{self.response_attr})
        return response.{self.response_attr}j
"""
        else:
            code += """
        self.sessionize(response)
        return response
"""
        return code



import textwrap

import click
import yaml

from botocraft.sync.models import BotocraftInterface

from .cli import cli


@cli.group(short_help="Inspect botocraft service definitions", name="data")
def data_group():
    pass


@data_group.command("models", short_help="List model definitions for a service")
@click.argument("service")
def data_list_models(service: str) -> None:
    """
    Print the model definitions for a service.

    Args:
        service: The name of the AWS service to print the manager definitions
            for

    Raises:
        click.ClickException: no service definition for the given service

    """
    interface = BotocraftInterface()
    interface.load()
    if service not in interface.services:
        msg = f'No service definition for AWS Service "{service}"'
        raise click.ClickException(msg)
    service_definition = interface.services[service]
    for model_name, model_def in service_definition.models.items():
        click.secho(model_name, fg="green")
        click.echo(textwrap.indent(yaml.dump(model_def.model_dump()), "  "))


@data_group.command("managers", short_help="List model definitions for a service")
@click.argument("service")
def data_list_managers(service: str) -> None:
    """
    Print the manager definitions for a service.

    Args:
        service: The name of the AWS service to print the manager definitions
            for

    Raises:
        click.ClickException: no service definition for the given service

    """
    interface = BotocraftInterface()
    interface.load()
    if service not in interface.services:
        msg = f'No service definition for AWS Service "{service}"'
        raise click.ClickException(msg)
    service_definition = interface.services[service]
    for mgr_name, mgr_def in service_definition.managers.items():
        click.secho(mgr_name, fg="green")
        click.echo(textwrap.indent(yaml.dump(mgr_def.model_dump()), "  "))



# This file is automatically generated by botocraft.  Do not edit directly.
# pylint: disable=anomalous-backslash-in-string,unsubscriptable-object,line-too-long,arguments-differ,arguments-renamed,unused-import,redefined-outer-name
# pyright: reportUnusedImport=false
# mypy: disable-error-code="index, override, assignment, union-attr, misc"
from datetime import datetime
from typing import Any, ClassVar, Dict, List, Literal, Optional, Type, cast

from pydantic import Field

from botocraft.mixins.tags import TagsDictMixin
from botocraft.services.common import Filter, Tag

from .abstract import (Boto3Model, Boto3ModelManager, PrimaryBoto3Model,
                       ReadonlyBoto3Model, ReadonlyBoto3ModelManager,
                       ReadonlyPrimaryBoto3Model)

# ===============
# Managers
# ===============


class DBInstanceManager(Boto3ModelManager):

    service_name: str = "rds"

    def create(
        self,
        model: "DBInstance",
        MasterUserPassword: Optional[str] = None,
        VpcSecurityGroupIds: Optional[List[str]] = None,
        DBSubnetGroupName: Optional[str] = None,
        DBParameterGroupName: Optional[str] = None,
        OptionGroupName: Optional[str] = None,
        TdeCredentialPassword: Optional[str] = None,
        Domain: Optional[str] = None,
        DomainFqdn: Optional[str] = None,
        DomainOu: Optional[str] = None,
        DomainAuthSecretArn: Optional[str] = None,
        DomainDnsIps: Optional[List[str]] = None,
        DomainIAMRoleName: Optional[str] = None,
        ManageMasterUserPassword: Optional[bool] = None,
        MasterUserSecretKmsKeyId: Optional[str] = None,
    ) -> "DBInstance":
        """
        Creates a new DB instance.

        Args:
            model: The :py:class:``DBInstance`` to create.

        Keyword Args:
            MasterUserPassword: The password for the master user.
            VpcSecurityGroupIds: A list of Amazon EC2 VPC security groups to associate
                with this DB instance.
            DBSubnetGroupName: A DB subnet group to associate with this DB instance.
            DBParameterGroupName: The name of the DB parameter group to associate with
                this DB instance. If you don't specify a value, then Amazon RDS uses the
                default DB parameter group for the specified DB engine and version.
            OptionGroupName: The option group to associate the DB instance with.
            TdeCredentialPassword: The password for the given ARN from the key store in
                order to access the device.
            Domain: The Active Directory directory ID to create the DB instance in.
                Currently, you can create only Db2, MySQL, Microsoft SQL Server, Oracle,
                and PostgreSQL DB instances in an Active Directory Domain.
            DomainFqdn: The fully qualified domain name (FQDN) of an Active Directory
                domain.
            DomainOu: The Active Directory organizational unit for your DB instance to
                join.
            DomainAuthSecretArn: The ARN for the Secrets Manager secret with the
                credentials for the user joining the domain.
            DomainDnsIps: The IPv4 DNS IP addresses of your primary and secondary
                Active Directory domain controllers.
            DomainIAMRoleName: The name of the IAM role to use when making API calls to
                the Directory Service.
            ManageMasterUserPassword: Specifies whether to manage the master user
                password with Amazon Web Services Secrets Manager.
            MasterUserSecretKmsKeyId: The Amazon Web Services KMS key identifier to
                encrypt a secret that is automatically generated and managed in Amazon Web
                Services Secrets Manager.
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            DBInstanceIdentifier=data.get("DBInstanceIdentifier"),
            DBInstanceClass=data.get("DBInstanceClass"),
            Engine=data.get("Engine"),
            DBName=data.get("DBName"),
            AllocatedStorage=data.get("AllocatedStorage"),
            MasterUsername=data.get("MasterUsername"),
            MasterUserPassword=self.serialize(MasterUserPassword),
            VpcSecurityGroupIds=self.serialize(VpcSecurityGroupIds),
            AvailabilityZone=data.get("AvailabilityZone"),
            DBSubnetGroupName=self.serialize(DBSubnetGroupName),
            PreferredMaintenanceWindow=data.get("PreferredMaintenanceWindow"),
            DBParameterGroupName=self.serialize(DBParameterGroupName),
            BackupRetentionPeriod=data.get("BackupRetentionPeriod"),
            PreferredBackupWindow=data.get("PreferredBackupWindow"),
            DbInstancePort=data.get("DbInstancePort"),
            MultiAZ=data.get("MultiAZ"),
            EngineVersion=data.get("EngineVersion"),
            AutoMinorVersionUpgrade=data.get("AutoMinorVersionUpgrade"),
            LicenseModel=data.get("LicenseModel"),
            Iops=data.get("Iops"),
            OptionGroupName=self.serialize(OptionGroupName),
            CharacterSetName=data.get("CharacterSetName"),
            NcharCharacterSetName=data.get("NcharCharacterSetName"),
            PubliclyAccessible=data.get("PubliclyAccessible"),
            TagList=data.get("TagList"),
            DBClusterIdentifier=data.get("DBClusterIdentifier"),
            StorageType=data.get("StorageType"),
            TdeCredentialArn=data.get("TdeCredentialArn"),
            TdeCredentialPassword=self.serialize(TdeCredentialPassword),
            StorageEncrypted=data.get("StorageEncrypted"),
            KmsKeyId=data.get("KmsKeyId"),
            Domain=self.serialize(Domain),
            DomainFqdn=self.serialize(DomainFqdn),
            DomainOu=self.serialize(DomainOu),
            DomainAuthSecretArn=self.serialize(DomainAuthSecretArn),
            DomainDnsIps=self.serialize(DomainDnsIps),
            CopyTagsToSnapshot=data.get("CopyTagsToSnapshot"),
            MonitoringInterval=data.get("MonitoringInterval"),
            MonitoringRoleArn=data.get("MonitoringRoleArn"),
            DomainIAMRoleName=self.serialize(DomainIAMRoleName),
            PromotionTier=data.get("PromotionTier"),
            Timezone=data.get("Timezone"),
            IAMDatabaseAuthenticationEnabled=data.get(
                "IAMDatabaseAuthenticationEnabled"
            ),
            PerformanceInsightsEnabled=data.get("PerformanceInsightsEnabled"),
            PerformanceInsightsKMSKeyId=data.get("PerformanceInsightsKMSKeyId"),
            PerformanceInsightsRetentionPeriod=data.get(
                "PerformanceInsightsRetentionPeriod"
            ),
            EnabledCloudwatchLogsExports=data.get("EnabledCloudwatchLogsExports"),
            ProcessorFeatures=data.get("ProcessorFeatures"),
            DeletionProtection=data.get("DeletionProtection"),
            MaxAllocatedStorage=data.get("MaxAllocatedStorage"),
            CustomerOwnedIpEnabled=data.get("CustomerOwnedIpEnabled"),
            CustomIamInstanceProfile=data.get("CustomIamInstanceProfile"),
            BackupTarget=data.get("BackupTarget"),
            NetworkType=data.get("NetworkType"),
            StorageThroughput=data.get("StorageThroughput"),
            ManageMasterUserPassword=self.serialize(ManageMasterUserPassword),
            MasterUserSecretKmsKeyId=self.serialize(MasterUserSecretKmsKeyId),
            CACertificateIdentifier=data.get("CACertificateIdentifier"),
            DBSystemId=data.get("DBSystemId"),
            DedicatedLogVolume=data.get("DedicatedLogVolume"),
            MultiTenant=data.get("MultiTenant"),
            EngineLifecycleSupport=data.get("EngineLifecycleSupport"),
        )
        _response = self.client.create_db_instance(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = CreateDBInstanceResult(**_response)

        self.sessionize(response.RDSDBInstance)
        return cast("DBInstance", response.RDSDBInstance)

    def update(
        self,
        model: "DBInstance",
        DBSubnetGroupName: Optional[str] = None,
        VpcSecurityGroupIds: Optional[List[str]] = None,
        ApplyImmediately: Optional[bool] = None,
        MasterUserPassword: Optional[str] = None,
        DBParameterGroupName: Optional[str] = None,
        AllowMajorVersionUpgrade: Optional[bool] = None,
        OptionGroupName: Optional[str] = None,
        NewDBInstanceIdentifier: Optional[str] = None,
        TdeCredentialPassword: Optional[str] = None,
        Domain: Optional[str] = None,
        DomainFqdn: Optional[str] = None,
        DomainOu: Optional[str] = None,
        DomainAuthSecretArn: Optional[str] = None,
        DomainDnsIps: Optional[List[str]] = None,
        DBPortNumber: Optional[int] = None,
        DomainIAMRoleName: Optional[str] = None,
        DisableDomain: Optional[bool] = None,
        CloudwatchLogsExportConfiguration: Optional[
            "CloudwatchLogsExportConfiguration"
        ] = None,
        UseDefaultProcessorFeatures: Optional[bool] = None,
        CertificateRotationRestart: Optional[bool] = None,
        ResumeFullAutomationModeMinutes: Optional[int] = None,
        ManageMasterUserPassword: Optional[bool] = None,
        RotateMasterUserPassword: Optional[bool] = None,
        MasterUserSecretKmsKeyId: Optional[str] = None,
    ) -> "DBInstance":
        """
        Modifies settings for a DB instance. You can change one or more
        database configuration parameters by specifying these parameters and
        the new values in the request. To learn what modifications you can make
        to your DB instance, call ``DescribeValidDBInstanceModifications``
        before you call ``ModifyDBInstance``.

        Args:
            model: The :py:class:``DBInstance`` to update.

        Keyword Args:
            DBSubnetGroupName: The new DB subnet group for the DB instance. You can use
                this parameter to move your DB instance to a different VPC. If your DB
                instance isn't in a VPC, you can also use this parameter to move your DB
                instance into a VPC. For more information, see `Working with a DB instance
                in a VPC <https://docs.aws.amazon. com/AmazonRDS/latest/UserGuide/USER_VPC.
                WorkingWithRDSInstanceinaVPC.html#USER_ VPC.Non-VPC2VPC>`_ in the *Amazon
                RDS User Guide*.
            VpcSecurityGroupIds: A list of Amazon EC2 VPC security groups to associate
                with this DB instance. This change is asynchronously applied as soon as
                possible.
            ApplyImmediately: Specifies whether the modifications in this request and
                any pending modifications are asynchronously applied as soon as possible,
                regardless of the ``PreferredMaintenanceWindow`` setting for the DB
                instance. By default, this parameter is disabled.
            MasterUserPassword: The new password for the master user.
            DBParameterGroupName: The name of the DB parameter group to apply to the DB
                instance.
            AllowMajorVersionUpgrade: Specifies whether major version upgrades are
                allowed. Changing this parameter doesn't result in an outage and the change
                is asynchronously applied as soon as possible.
            OptionGroupName: The option group to associate the DB instance with.
            NewDBInstanceIdentifier: The new identifier for the DB instance when
                renaming a DB instance. When you change the DB instance identifier, an
                instance reboot occurs immediately if you enable ``ApplyImmediately``, or
                will occur during the next maintenance window if you disable
                ``ApplyImmediately``. This value is stored as a lowercase string.
            TdeCredentialPassword: The password for the given ARN from the key store in
                order to access the device.
            Domain: The Active Directory directory ID to move the DB instance to.
                Specify ``none`` to remove the instance from its current domain. You must
                create the domain before this operation. Currently, you can create only
                Db2, MySQL, Microsoft SQL Server, Oracle, and PostgreSQL DB instances in an
                Active Directory Domain.
            DomainFqdn: The fully qualified domain name (FQDN) of an Active Directory
                domain.
            DomainOu: The Active Directory organizational unit for your DB instance to
                join.
            DomainAuthSecretArn: The ARN for the Secrets Manager secret with the
                credentials for the user joining the domain.
            DomainDnsIps: The IPv4 DNS IP addresses of your primary and secondary
                Active Directory domain controllers.
            DBPortNumber: The port number on which the database accepts connections.
            DomainIAMRoleName: The name of the IAM role to use when making API calls to
                the Directory Service.
            DisableDomain: Specifies whether to remove the DB instance from the Active
                Directory domain.
            CloudwatchLogsExportConfiguration: The log types to be enabled for export
                to CloudWatch Logs for a specific DB instance.
            UseDefaultProcessorFeatures: Specifies whether the DB instance class of the
                DB instance uses its default processor features.
            CertificateRotationRestart: Specifies whether the DB instance is restarted
                when you rotate your SSL/TLS certificate.
            ResumeFullAutomationModeMinutes: The number of minutes to pause the
                automation. When the time period ends, RDS Custom resumes full automation.
            ManageMasterUserPassword: Specifies whether to manage the master user
                password with Amazon Web Services Secrets Manager.
            RotateMasterUserPassword: Specifies whether to rotate the secret managed by
                Amazon Web Services Secrets Manager for the master user password.
            MasterUserSecretKmsKeyId: The Amazon Web Services KMS key identifier to
                encrypt a secret that is automatically generated and managed in Amazon Web
                Services Secrets Manager.
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            DBInstanceIdentifier=data.get("DBInstanceIdentifier"),
            AllocatedStorage=data.get("AllocatedStorage"),
            DBInstanceClass=data.get("DBInstanceClass"),
            DBSubnetGroupName=self.serialize(DBSubnetGroupName),
            VpcSecurityGroupIds=self.serialize(VpcSecurityGroupIds),
            ApplyImmediately=self.serialize(ApplyImmediately),
            MasterUserPassword=self.serialize(MasterUserPassword),
            DBParameterGroupName=self.serialize(DBParameterGroupName),
            BackupRetentionPeriod=data.get("BackupRetentionPeriod"),
            PreferredBackupWindow=data.get("PreferredBackupWindow"),
            PreferredMaintenanceWindow=data.get("PreferredMaintenanceWindow"),
            MultiAZ=data.get("MultiAZ"),
            EngineVersion=data.get("EngineVersion"),
            AllowMajorVersionUpgrade=self.serialize(AllowMajorVersionUpgrade),
            AutoMinorVersionUpgrade=data.get("AutoMinorVersionUpgrade"),
            LicenseModel=data.get("LicenseModel"),
            Iops=data.get("Iops"),
            OptionGroupName=self.serialize(OptionGroupName),
            NewDBInstanceIdentifier=self.serialize(NewDBInstanceIdentifier),
            StorageType=data.get("StorageType"),
            TdeCredentialArn=data.get("TdeCredentialArn"),
            TdeCredentialPassword=self.serialize(TdeCredentialPassword),
            CACertificateIdentifier=data.get("CACertificateIdentifier"),
            Domain=self.serialize(Domain),
            DomainFqdn=self.serialize(DomainFqdn),
            DomainOu=self.serialize(DomainOu),
            DomainAuthSecretArn=self.serialize(DomainAuthSecretArn),
            DomainDnsIps=self.serialize(DomainDnsIps),
            CopyTagsToSnapshot=data.get("CopyTagsToSnapshot"),
            MonitoringInterval=data.get("MonitoringInterval"),
            DBPortNumber=self.serialize(DBPortNumber),
            PubliclyAccessible=data.get("PubliclyAccessible"),
            MonitoringRoleArn=data.get("MonitoringRoleArn"),
            DomainIAMRoleName=self.serialize(DomainIAMRoleName),
            DisableDomain=self.serialize(DisableDomain),
            PromotionTier=data.get("PromotionTier"),
            IAMDatabaseAuthenticationEnabled=data.get(
                "IAMDatabaseAuthenticationEnabled"
            ),
            PerformanceInsightsEnabled=data.get("PerformanceInsightsEnabled"),
            PerformanceInsightsKMSKeyId=data.get("PerformanceInsightsKMSKeyId"),
            PerformanceInsightsRetentionPeriod=data.get(
                "PerformanceInsightsRetentionPeriod"
            ),
            CloudwatchLogsExportConfiguration=self.serialize(
                CloudwatchLogsExportConfiguration
            ),
            ProcessorFeatures=data.get("ProcessorFeatures"),
            UseDefaultProcessorFeatures=self.serialize(UseDefaultProcessorFeatures),
            DeletionProtection=data.get("DeletionProtection"),
            MaxAllocatedStorage=data.get("MaxAllocatedStorage"),
            CertificateRotationRestart=self.serialize(CertificateRotationRestart),
            ReplicaMode=data.get("ReplicaMode"),
            CustomerOwnedIpEnabled=data.get("CustomerOwnedIpEnabled"),
            AwsBackupRecoveryPointArn=data.get("AwsBackupRecoveryPointArn"),
            AutomationMode=data.get("AutomationMode"),
            ResumeFullAutomationModeMinutes=self.serialize(
                ResumeFullAutomationModeMinutes
            ),
            NetworkType=data.get("NetworkType"),
            StorageThroughput=data.get("StorageThroughput"),
            ManageMasterUserPassword=self.serialize(ManageMasterUserPassword),
            RotateMasterUserPassword=self.serialize(RotateMasterUserPassword),
            MasterUserSecretKmsKeyId=self.serialize(MasterUserSecretKmsKeyId),
            Engine=data.get("Engine"),
            DedicatedLogVolume=data.get("DedicatedLogVolume"),
            MultiTenant=data.get("MultiTenant"),
        )
        _response = self.client.modify_db_instance(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = ModifyDBInstanceResult(**_response)

        self.sessionize(response.RDSDBInstance)
        return cast("DBInstance", response.RDSDBInstance)

    def delete(
        self,
        DBInstanceIdentifier: str,
        *,
        SkipFinalSnapshot: Optional[bool] = None,
        FinalDBSnapshotIdentifier: Optional[str] = None,
        DeleteAutomatedBackups: Optional[bool] = None
    ) -> "DBInstance":
        """
        Deletes a previously provisioned DB instance. When you delete a DB
        instance, all automated backups for that instance are deleted and can't
        be recovered. However, manual DB snapshots of the DB instance aren't
        deleted.

        Args:
            DBInstanceIdentifier: The DB instance identifier for the DB instance to be
                deleted. This parameter isn't case-sensitive.

        Keyword Args:
            SkipFinalSnapshot: Specifies whether to skip the creation of a final DB
                snapshot before deleting the instance. If you enable this parameter, RDS
                doesn't create a DB snapshot. If you don't enable this parameter, RDS
                creates a DB snapshot before the DB instance is deleted. By default, skip
                isn't enabled, and the DB snapshot is created.
            FinalDBSnapshotIdentifier: The ``DBSnapshotIdentifier`` of the new
                ``DBSnapshot`` created when the ``SkipFinalSnapshot`` parameter is
                disabled.
            DeleteAutomatedBackups: Specifies whether to remove automated backups
                immediately after the DB instance is deleted. This parameter isn't case-
                sensitive. The default is to remove automated backups immediately after the
                DB instance is deleted.
        """
        args: Dict[str, Any] = dict(
            DBInstanceIdentifier=self.serialize(DBInstanceIdentifier),
            SkipFinalSnapshot=self.serialize(SkipFinalSnapshot),
            FinalDBSnapshotIdentifier=self.serialize(FinalDBSnapshotIdentifier),
            DeleteAutomatedBackups=self.serialize(DeleteAutomatedBackups),
        )
        _response = self.client.delete_db_instance(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = DeleteDBInstanceResult(**_response)
        return cast(DBInstance, response.RDSDBInstance)

    def get(
        self, DBInstanceIdentifier: str, *, Filters: Optional[List[Filter]] = None
    ) -> Optional["DBInstance"]:
        """
        Describes provisioned RDS instances. This API supports pagination.

        Args:
            DBInstanceIdentifier: The user-supplied instance identifier or the Amazon
                Resource Name (ARN) of the DB instance. If this parameter is specified,
                information from only the specific DB instance is returned. This parameter
                isn't case-sensitive.

        Keyword Args:
            Filters: A filter that specifies one or more DB instances to describe.
        """
        args: Dict[str, Any] = dict(
            DBInstanceIdentifier=self.serialize(DBInstanceIdentifier),
            Filters=self.serialize(Filters),
        )
        _response = self.client.describe_db_instances(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = DBInstanceMessage(**_response)

        if response and response.DBInstances:
            self.sessionize(response.DBInstances[0])
            return response.DBInstances[0]
        return None

    def list(
        self,
        *,
        DBInstanceIdentifier: Optional[str] = None,
        Filters: Optional[List[Filter]] = None
    ) -> List["DBInstance"]:
        """
        Describes provisioned RDS instances. This API supports pagination.

        Keyword Args:
            DBInstanceIdentifier: The user-supplied instance identifier or the Amazon
                Resource Name (ARN) of the DB instance. If this parameter is specified,
                information from only the specific DB instance is returned. This parameter
                isn't case-sensitive.
            Filters: A filter that specifies one or more DB instances to describe.
        """
        paginator = self.client.get_paginator("describe_db_instances")
        args: Dict[str, Any] = dict(
            DBInstanceIdentifier=self.serialize(DBInstanceIdentifier),
            Filters=self.serialize(Filters),
        )
        response_iterator = paginator.paginate(
            **{k: v for k, v in args.items() if v is not None}
        )
        results: List["DBInstance"] = []
        for _response in response_iterator:
            response = DBInstanceMessage(**_response)
            if response.DBInstances:
                results.extend(response.DBInstances)
            else:
                break
        self.sessionize(results)
        return results


# ==============
# Service Models
# ==============


class RDSEndpoint(Boto3Model):
    """
    The connection endpoint for the DB instance.

    The endpoint might not be shown for instances with the status of ``creating``.
    """

    Address: Optional[str] = None
    """
    Specifies the DNS address of the DB instance.
    """
    Port: Optional[int] = None
    """
    Specifies the port that the database engine is listening on.
    """
    HostedZoneId: Optional[str] = None
    """
    Specifies the ID that Amazon Route 53 assigns when you create a hosted
    zone.
    """


class DBSecurityGroupMembership(Boto3Model):
    """This data type is used as a response element in the following actions:

    * ``ModifyDBInstance``
    * ``RebootDBInstance``
    * ``RestoreDBInstanceFromDBSnapshot``
    * ``RestoreDBInstanceToPointInTime``
    """

    DBSecurityGroupName: Optional[str] = None
    """
    The name of the DB security group.
    """
    Status: Optional[str] = None
    """
    The status of the DB security group.
    """


class VpcSecurityGroupMembership(Boto3Model):
    """
    This data type is used as a response element for queries on VPC security
    group membership.
    """

    VpcSecurityGroupId: Optional[str] = None
    """
    The name of the VPC security group.
    """
    Status: Optional[str] = None
    """
    The membership status of the VPC security group.
    """


class DBParameterGroupStatus(Boto3Model):
    """
    The status of the DB parameter group.

    This data type is used as a response element in the following actions:

    * ``CreateDBInstance``
    * ``CreateDBInstanceReadReplica``
    * ``DeleteDBInstance``
    * ``ModifyDBInstance``
    * ``RebootDBInstance``
    * ``RestoreDBInstanceFromDBSnapshot``
    """

    DBParameterGroupName: Optional[str] = None
    """
    The name of the DB parameter group.
    """
    ParameterApplyStatus: Optional[str] = None
    """
    The status of parameter updates.

    Valid values are:
    """


class RDSAvailabilityZone(Boto3Model):
    """
    Contains Availability Zone information.

    This data type is used as an element in the ``OrderableDBInstanceOption`` data
    type.
    """

    Name: Optional[str] = None
    """
    The name of the Availability Zone.
    """


class Outpost(Boto3Model):
    """
    If the subnet is associated with an Outpost, this value specifies the
    Outpost.

    For more information about RDS on Outposts, see
    `Amazon RDS on Amazon Web Services Outposts <https://docs.aws.amazon.com/AmazonRDS/latest/UserGuide/rds-
    on-outposts.html>`_ in the *Amazon RDS User Guide.*
    """

    Arn: Optional[str] = None
    """
    The Amazon Resource Name (ARN) of the Outpost.
    """


class RDSSubnet(Boto3Model):
    """
    This data type is used as a response element for the
    ``DescribeDBSubnetGroups`` operation.
    """

    SubnetIdentifier: Optional[str] = None
    """
    The identifier of the subnet.
    """
    SubnetAvailabilityZone: Optional[RDSAvailabilityZone] = None
    """
    Contains Availability Zone information.
    """
    SubnetOutpost: Optional[Outpost] = None
    """
    If the subnet is associated with an Outpost, this value specifies the
    Outpost.
    """
    SubnetStatus: Optional[str] = None
    """
    The status of the subnet.
    """


class RDSDBSubnetGroup(Boto3Model):
    """
    Information about the subnet group associated with the DB instance,
    including the name, description, and subnets in the subnet group.
    """

    DBSubnetGroupName: Optional[str] = None
    """
    The name of the DB subnet group.
    """
    DBSubnetGroupDescription: Optional[str] = None
    """
    Provides the description of the DB subnet group.
    """
    VpcId: Optional[str] = None
    """
    Provides the VpcId of the DB subnet group.
    """
    SubnetGroupStatus: Optional[str] = None
    """
    Provides the status of the DB subnet group.
    """
    Subnets: Optional[List["RDSSubnet"]] = None
    """
    Contains a list of ``Subnet`` elements.
    """
    DBSubnetGroupArn: Optional[str] = None
    """
    The Amazon Resource Name (ARN) for the DB subnet group.
    """
    SupportedNetworkTypes: Optional[List[str]] = None
    """
    The network type of the DB subnet group.
    """


class RDSPendingCloudwatchLogsExports(Boto3Model):
    """
    A list of the log types whose configuration is still pending.

    In other words, these log types are in the process of being activated or
    deactivated.
    """

    LogTypesToEnable: Optional[List[str]] = None
    """
    Log types that are in the process of being deactivated.

    After they are deactivated, these log types aren't exported to CloudWatch
    Logs.
    """
    LogTypesToDisable: Optional[List[str]] = None
    """
    Log types that are in the process of being enabled.

    After they are enabled, these log types are exported to CloudWatch Logs.
    """


class ProcessorFeature(Boto3Model):
    """
    Contains the processor features of a DB instance class.

    To specify the number of CPU cores, use the ``coreCount`` feature name for the
    ``Name`` parameter. To specify the number of threads per core, use the
    ``threadsPerCore`` feature name for the ``Name`` parameter.

    You can set the processor features of the DB instance class for a DB instance
    when you call one of the following actions:

    * ``CreateDBInstance``
    * ``ModifyDBInstance``
    * ``RestoreDBInstanceFromDBSnapshot``
    * ``RestoreDBInstanceFromS3``
    * ``RestoreDBInstanceToPointInTime``

    You can view the valid processor values for a particular instance class by
    calling the ``DescribeOrderableDBInstanceOptions`` action and specifying the
    instance class for the ``DBInstanceClass`` parameter.

    In addition, you can use the following actions for DB instance class processor
    information:

    * ``DescribeDBInstances``
    * ``DescribeDBSnapshots``
    * ``DescribeValidDBInstanceModifications``

    If you call ``DescribeDBInstances``, ``ProcessorFeature`` returns non-null
    values only if the following conditions are met:

    * You are accessing an Oracle DB instance.
    * Your Oracle DB instance class supports configuring the number of CPU cores
      and threads per core.
    * The current number CPU cores and threads is set to a non-default value.

    For more information, see  `Configuring the processor for a DB instance class
    in RDS for Oracle <https://docs.aws.amazon.com/AmazonRDS/latest/UserGuide/Conce
    pts.DBInstanceClass.html#USER_ConfigureProcessor>`_ in the *Amazon RDS User
    Guide.*
    """

    Name: Optional[str] = None
    """
    The name of the processor feature.

    Valid names are ``coreCount`` and
    ``threadsPerCore``.
    """
    Value: Optional[str] = None
    """
    The value of a processor feature.
    """


class RDSPendingModifiedValues(Boto3Model):
    """
    Information about pending changes to the DB instance.

    This information is returned only when there are pending changes. Specific
    changes are identified by subelements.
    """

    DBInstanceClass: Optional[str] = None
    """
    The name of the compute and memory capacity class for the DB instance.
    """
    AllocatedStorage: Optional[int] = None
    """
    The allocated storage size for the DB instance specified in gibibytes
    (GiB).
    """
    MasterUserPassword: Optional[str] = None
    """
    The master credentials for the DB instance.
    """
    Port: Optional[int] = None
    """
    The port for the DB instance.
    """
    BackupRetentionPeriod: Optional[int] = None
    """
    The number of days for which automated backups are retained.
    """
    MultiAZ: Optional[bool] = None
    """
    Indicates whether the Single-AZ DB instance will change to a Multi-AZ
    deployment.
    """
    EngineVersion: Optional[str] = None
    """
    The database engine version.
    """
    LicenseModel: Optional[str] = None
    """
    The license model for the DB instance.
    """
    Iops: Optional[int] = None
    """
    The Provisioned IOPS value for the DB instance.
    """
    DBInstanceIdentifier: Optional[str] = None
    """
    The database identifier for the DB instance.
    """
    StorageType: Optional[str] = None
    """
    The storage type of the DB instance.
    """
    CACertificateIdentifier: Optional[str] = None
    """
    The identifier of the CA certificate for the DB instance.
    """
    DBSubnetGroupName: Optional[str] = None
    """
    The DB subnet group for the DB instance.
    """
    PendingCloudwatchLogsExports: Optional[RDSPendingCloudwatchLogsExports] = None
    """
    A list of the log types whose configuration is still pending.

    In other words, these log types are in the process of being activated or
    deactivated.
    """
    ProcessorFeatures: Optional[List["ProcessorFeature"]] = None
    """
    The number of CPU cores and the number of threads per core for the DB
    instance class of the DB instance.
    """
    IAMDatabaseAuthenticationEnabled: Optional[bool] = None
    """
    Indicates whether mapping of Amazon Web Services Identity and Access
    Management (IAM) accounts to database accounts is enabled.
    """
    AutomationMode: Optional[Literal["full", "all-paused"]] = None
    """
The automation mode of the RDS Custom DB instance: ``full`` or ``all-paused``.
If ``full``, the DB instance automates monitoring and instance recovery. If
``all-paused``, the instance pauses automation for the duration set by
``--resume-full-automation-mode-minutes``.
    """
    ResumeFullAutomationModeTime: Optional[datetime] = None
    """
    The number of minutes to pause the automation.

    When the time period ends, RDS Custom resumes full automation. The minimum
    value is 60 (default). The maximum value is 1,440.
    """
    StorageThroughput: Optional[int] = None
    """
    The storage throughput of the DB instance.
    """
    Engine: Optional[str] = None
    """
    The database engine of the DB instance.
    """
    DedicatedLogVolume: Optional[bool] = None
    """
    Indicates whether the DB instance has a dedicated log volume (DLV)
    enabled.>
    """
    MultiTenant: Optional[bool] = None
    """
    Indicates whether the DB instance will change to the multi-tenant
    configuration (TRUE) or the single-tenant configuration (FALSE).
    """


class OptionGroupMembership(Boto3Model):
    """
    Provides information on the option groups the DB instance is a member of.
    """

    OptionGroupName: Optional[str] = None
    """
    The name of the option group that the instance belongs to.
    """
    Status: Optional[str] = None
    """
    The status of the DB instance's option group membership.

    Valid values are:
    ``in-sync``, ``pending-apply``, ``pending-removal``, ``pending-maintenance-
    apply``, ``pending-maintenance-removal``, ``applying``, ``removing``, and
    ``failed``.
    """


class DBInstanceStatusInfo(Boto3Model):
    """
    Provides a list of status information for a DB instance.
    """

    StatusType: Optional[str] = None
    """
    This value is currently "read replication.".
    """
    Normal: Optional[bool] = None
    """
    Indicates whether the instance is operating normally (TRUE) or is in an
    error state (FALSE).
    """
    Status: Optional[str] = None
    """
    The status of the DB instance.

    For a StatusType of read replica, the values can be replicating,
    replication stop point set, replication stop point reached, error, stopped,
    or terminated.
    """
    Message: Optional[str] = None
    """
    Details of the error if there is an error for the instance.

    If the instance isn't in an error state, this value is blank.
    """


class DomainMembership(Boto3Model):
    """
    An Active Directory Domain membership record associated with the DB
    instance or cluster.
    """

    Domain: Optional[str] = None
    """
    The identifier of the Active Directory Domain.
    """
    Status: Optional[str] = None
    """
    The status of the Active Directory Domain membership for the DB instance or
    cluster.

    Values include ``joined``, ``pending-join``, ``failed``, and so on.
    """
    FQDN: Optional[str] = None
    """
    The fully qualified domain name (FQDN) of the Active Directory Domain.
    """
    IAMRoleName: Optional[str] = None
    """
    The name of the IAM role used when making API calls to the Directory
    Service.
    """
    OU: Optional[str] = None
    """
    The Active Directory organizational unit for the DB instance or cluster.
    """
    AuthSecretArn: Optional[str] = None
    """
    The ARN for the Secrets Manager secret with the credentials for the user
    that's a member of the domain.
    """
    DnsIps: Optional[List[str]] = None
    """
    The IPv4 DNS IP addresses of the primary and secondary Active Directory
    domain controllers.
    """


class DBInstanceRole(Boto3Model):
    """
    Information about an Amazon Web Services Identity and Access Management
    (IAM) role that is associated with a DB instance.
    """

    RoleArn: Optional[str] = None
    """
    The Amazon Resource Name (ARN) of the IAM role that is associated with the
    DB instance.
    """
    FeatureName: Optional[str] = None
    """
    The name of the feature associated with the Amazon Web Services Identity
    and Access Management (IAM) role.

    For information about supported feature names,
    see ``DBEngineVersion``.
    """
    Status: Optional[str] = None
    """
    Information about the state of association between the IAM role and the DB
    instance.

    The Status property returns one of the following values:
    """


class DBInstanceAutomatedBackupsReplication(Boto3Model):
    """
    Automated backups of a DB instance replicated to another Amazon Web
    Services Region.

    They consist of system backups, transaction logs, and database instance
    properties.
    """

    DBInstanceAutomatedBackupsArn: Optional[str] = None
    """
    The Amazon Resource Name (ARN) of the replicated automated backups.
    """


class RDSMasterUserSecret(Boto3Model):
    """
    The secret managed by RDS in Amazon Web Services Secrets Manager for the
    master user password.

    For more information, see
    `Password management with Amazon Web Services Secrets Manager <https://docs.aws.amazon.com/AmazonRDS/latest/UserGuide/rds-secrets-
    manager.html>`_ in the *Amazon RDS User Guide.*
    """

    SecretArn: Optional[str] = None
    """
    The Amazon Resource Name (ARN) of the secret.
    """
    SecretStatus: Optional[str] = None
    """
    The status of the secret.
    """
    KmsKeyId: Optional[str] = None
    """
    The Amazon Web Services KMS key identifier that is used to encrypt the
    secret.
    """


class RDSCertificateDetails(Boto3Model):
    """
    The details of the DB instance's server certificate.
    """

    CAIdentifier: Optional[str] = None
    """
    The CA identifier of the CA certificate used for the DB instance's server
    certificate.
    """
    ValidTill: Optional[datetime] = None
    """
    The expiration date of the DB instance’s server certificate.
    """


class DBInstance(PrimaryBoto3Model):
    """
    Contains the details of an Amazon RDS DB instance.

    This data type is used as a response element in the operations
    ``CreateDBInstance``, ``CreateDBInstanceReadReplica``, ``DeleteDBInstance``,
    ``DescribeDBInstances``, ``ModifyDBInstance``, ``PromoteReadReplica``,
    ``RebootDBInstance``, ``RestoreDBInstanceFromDBSnapshot``,
    ``RestoreDBInstanceFromS3``, ``RestoreDBInstanceToPointInTime``,
    ``StartDBInstance``, and ``StopDBInstance``.
    """

    manager_class: ClassVar[Type[Boto3ModelManager]] = DBInstanceManager

    DBInstanceIdentifier: str
    """
    The user-supplied database identifier.

    This identifier is the unique key that identifies a DB instance.
    """
    DBInstanceClass: str
    """
    The name of the compute and memory capacity class of the DB instance.
    """
    MultiAZ: Optional[bool]
    """
    Indicates whether the DB instance is a Multi-AZ deployment.

    This setting doesn't apply to RDS Custom DB instances.
    """
    Engine: str
    """
    The database engine used for this DB instance.
    """
    EngineVersion: str
    """
    The version of the database engine.
    """
    AutoMinorVersionUpgrade: Optional[bool] = True
    """
    Indicates whether minor version patches are applied automatically.
    """
    PubliclyAccessible: Optional[bool]
    """
    Indicates whether the DB instance is publicly accessible.
    """
    StorageEncrypted: Optional[bool] = True
    """
    Indicates whether the DB instance is encrypted.
    """
    BackupRetentionPeriod: Optional[int] = 1
    """
    The number of days for which automatic DB snapshots are retained.
    """
    DbInstancePort: Optional[int] = None
    """
    The port that the DB instance listens on.

    If the DB instance is part of a DB cluster, this can be a different port
    than the DB cluster port.
    """
    IAMDatabaseAuthenticationEnabled: Optional[bool] = None
    """
    Indicates whether mapping of Amazon Web Services Identity and Access
    Management (IAM) accounts to database accounts is enabled for the DB
    instance.
    """
    EnabledCloudwatchLogsExports: Optional[List[str]] = None
    """
    A list of log types that this DB instance is configured to export to
    CloudWatch Logs.
    """
    TagList: Optional[List[Tag]] = None
    """
    A list of tags.
    """
    CustomerOwnedIpEnabled: Optional[bool] = None
    """
    Indicates whether a customer-owned IP address (CoIP) is enabled for an RDS
    on Outposts DB instance.
    """
    PerformanceInsightsEnabled: Optional[bool] = None
    """
    Indicates whether Performance Insights is enabled for the DB instance.
    """
    DBInstanceStatus: str = Field(default=None, frozen=True)
    """
    The current state of this database.
    """
    AutomaticRestartTime: datetime = Field(default=None, frozen=True)
    """
    The time when a stopped DB instance is restarted automatically.
    """
    MasterUsername: Optional[str] = None
    """
    The master username for the DB instance.
    """
    DBName: Optional[str] = None
    """
    The initial database name that you provided (if required) when you created
    the DB instance.

    This name is returned for the life of your DB instance. For an RDS for
    Oracle CDB instance, the name identifies the PDB rather than the CDB.
    """
    Endpoint: RDSEndpoint = Field(default=None, frozen=True)
    """
    The connection endpoint for the DB instance.
    """
    AllocatedStorage: Optional[int] = None
    """
    The amount of storage in gibibytes (GiB) allocated for the DB instance.
    """
    InstanceCreateTime: datetime = Field(default=None, frozen=True)
    """
    The date and time when the DB instance was created.
    """
    PreferredBackupWindow: Optional[str] = None
    """
    The daily time range during which automated backups are created if
    automated backups are enabled, as determined by the
    ``BackupRetentionPeriod``.
    """
    DBSecurityGroups: Optional[List["DBSecurityGroupMembership"]] = None
    """
    A list of DB security group elements containing ``DBSecurityGroup.Name``
    and ``DBSecurityGroup.Status`` subelements.
    """
    VpcSecurityGroups: List["VpcSecurityGroupMembership"] = Field(
        default=None, frozen=True
    )
    """
    The list of Amazon EC2 VPC security groups that the DB instance belongs to.
    """
    DBParameterGroups: List["DBParameterGroupStatus"] = Field(default=None, frozen=True)
    """
    The list of DB parameter groups applied to this DB instance.
    """
    AvailabilityZone: Optional[str] = None
    """
    The name of the Availability Zone where the DB instance is located.
    """
    DBSubnetGroup: RDSDBSubnetGroup = Field(default=None, frozen=True)
    """
    Information about the subnet group associated with the DB instance,
    including the name, description, and subnets in the subnet group.
    """
    PreferredMaintenanceWindow: Optional[str] = None
    """
    The weekly time range during which system maintenance can occur, in
    Universal Coordinated Time (UTC).
    """
    PendingModifiedValues: RDSPendingModifiedValues = Field(default=None, frozen=True)
    """
    Information about pending changes to the DB instance.

    This information is returned only when there are pending changes. Specific
    changes are identified by subelements.
    """
    LatestRestorableTime: datetime = Field(default=None, frozen=True)
    """
    The latest time to which a database in this DB instance can be restored
    with point-in-time restore.
    """
    ReadReplicaSourceDBInstanceIdentifier: str = Field(default=None, frozen=True)
    """
    The identifier of the source DB instance if this DB instance is a read
    replica.
    """
    ReadReplicaDBInstanceIdentifiers: List[str] = Field(default=None, frozen=True)
    """
    The identifiers of the read replicas associated with this DB instance.
    """
    ReadReplicaDBClusterIdentifiers: List[str] = Field(default=None, frozen=True)
    """
    The identifiers of Aurora DB clusters to which the RDS DB instance is
    replicated as a read replica.

    For example, when you create an Aurora read replica of an RDS for MySQL DB
    instance, the Aurora MySQL DB cluster for the Aurora read replica is shown.
    This output doesn't contain information about cross-Region Aurora read
    replicas.
    """
    ReplicaMode: Optional[Literal["open-read-only", "mounted"]] = None
    """
    The open mode of an Oracle read replica.

    The default is ``open-read-only``. For
    more information, see `Working with Oracle Read Replicas for Amazon
    RDS <https://docs.aws.amazon.com/AmazonRDS/latest/UserGuide/oracle-read-
    replicas.html>`_ in the *Amazon RDS User Guide*.
    """
    LicenseModel: Optional[str] = None
    """
    The license model information for this DB instance.

    This setting doesn't apply to Amazon Aurora or RDS Custom DB instances.
    """
    Iops: Optional[int] = None
    """
    The Provisioned IOPS (I/O operations per second) value for the DB instance.
    """
    OptionGroupMemberships: List["OptionGroupMembership"] = Field(
        default=None, frozen=True
    )
    """
    The list of option group memberships for this DB instance.
    """
    CharacterSetName: Optional[str] = None
    """
    If present, specifies the name of the character set that this instance is
    associated with.
    """
    NcharCharacterSetName: Optional[str] = None
    """
    The name of the NCHAR character set for the Oracle DB instance.

    This character set specifies the Unicode encoding for data stored in table
    columns of type NCHAR, NCLOB, or NVARCHAR2.
    """
    SecondaryAvailabilityZone: str = Field(default=None, frozen=True)
    """
    If present, specifies the name of the secondary Availability Zone for a DB
    instance with multi-AZ support.
    """
    StatusInfos: List["DBInstanceStatusInfo"] = Field(default=None, frozen=True)
    """
    The status of a read replica.

    If the DB instance isn't a read replica, the value is blank.
    """
    StorageType: Optional[str] = None
    """
    The storage type associated with the DB instance.
    """
    TdeCredentialArn: Optional[str] = None
    """
    The ARN from the key store with which the instance is associated for TDE
    encryption.
    """
    DBClusterIdentifier: Optional[str] = None
    """
    If the DB instance is a member of a DB cluster, indicates the name of the
    DB cluster that the DB instance is a member of.
    """
    KmsKeyId: Optional[str] = None
    """
    If ``StorageEncrypted`` is enabled, the Amazon Web Services KMS key
    identifier for the encrypted DB instance.
    """
    DbiResourceId: str = Field(default=None, frozen=True)
    """
    The Amazon Web Services Region-unique, immutable identifier for the DB
    instance.

    This identifier is found in Amazon Web Services CloudTrail log entries
    whenever the Amazon Web Services KMS key for the DB instance is accessed.
    """
    CACertificateIdentifier: Optional[str] = None
    """
    The identifier of the CA certificate for this DB instance.
    """
    DomainMemberships: List["DomainMembership"] = Field(default=None, frozen=True)
    """
    The Active Directory Domain membership records associated with the DB
    instance.
    """
    CopyTagsToSnapshot: Optional[bool] = None
    """
    Indicates whether tags are copied from the DB instance to snapshots of the
    DB instance.
    """
    MonitoringInterval: Optional[int] = None
    """
    The interval, in seconds, between points when Enhanced Monitoring metrics
    are collected for the DB instance.
    """
    EnhancedMonitoringResourceArn: str = Field(default=None, frozen=True)
    """
    The Amazon Resource Name (ARN) of the Amazon CloudWatch Logs log stream
    that receives the Enhanced Monitoring metrics data for the DB instance.
    """
    MonitoringRoleArn: Optional[str] = None
    """
    The ARN for the IAM role that permits RDS to send Enhanced Monitoring
    metrics to Amazon CloudWatch Logs.
    """
    PromotionTier: Optional[int] = None
    """
    The order of priority in which an Aurora Replica is promoted to the primary
    instance after a failure of the existing primary instance.

    For more
    information, see  `Fault Tolerance for an Aurora DB Cluster <https://docs.aws.a
    mazon.com/AmazonRDS/latest/AuroraUserGuide/Concepts.AuroraHighAvailability.html
    #Aurora.Managing.FaultTolerance>`_ in the *Amazon Aurora User Guide*.
    """
    DBInstanceArn: str = Field(default=None, frozen=True)
    """
    The Amazon Resource Name (ARN) for the DB instance.
    """
    Timezone: Optional[str] = None
    """
    The time zone of the DB instance.

    In most cases, the ``Timezone`` element is
    empty. ``Timezone`` content appears only for RDS for Db2 and RDS for SQL Server
    DB instances that were created with a time zone specified.
    """
    PerformanceInsightsKMSKeyId: Optional[str] = None
    """
    The Amazon Web Services KMS key identifier for encryption of Performance
    Insights data.
    """
    PerformanceInsightsRetentionPeriod: Optional[int] = None
    """
    The number of days to retain Performance Insights data.
    """
    ProcessorFeatures: Optional[List["ProcessorFeature"]] = None
    """
    The number of CPU cores and the number of threads per core for the DB
    instance class of the DB instance.
    """
    DeletionProtection: Optional[bool] = None
    """
    Indicates whether the DB instance has deletion protection enabled.

    The database can't be deleted when deletion protection is enabled. For more
    information, see
    `Deleting a DB Instance <https://docs.aws.amazon.com/AmazonRDS/latest/UserGuide
    /USER_DeleteInstance.html>`_.
    """
    AssociatedRoles: List["DBInstanceRole"] = Field(default=None, frozen=True)
    """
    The Amazon Web Services Identity and Access Management (IAM) roles
    associated with the DB instance.
    """
    ListenerEndpoint: RDSEndpoint = Field(default=None, frozen=True)
    """
    The listener connection endpoint for SQL Server Always On.
    """
    MaxAllocatedStorage: Optional[int] = None
    """
    The upper limit in gibibytes (GiB) to which Amazon RDS can automatically
    scale the storage of the DB instance.
    """
    DBInstanceAutomatedBackupsReplications: List[
        "DBInstanceAutomatedBackupsReplication"
    ] = Field(default=None, frozen=True)
    """
    The list of replicated automated backups associated with the DB instance.
    """
    AwsBackupRecoveryPointArn: Optional[str] = None
    """
    The Amazon Resource Name (ARN) of the recovery point in Amazon Web Services
    Backup.
    """
    ActivityStreamStatus: Literal["stopped", "starting", "started", "stopping"] = Field(
        default=None, frozen=True
    )
    """
    The status of the database activity stream.
    """
    ActivityStreamKmsKeyId: str = Field(default=None, frozen=True)
    """
    The Amazon Web Services KMS key identifier used for encrypting messages in
    the database activity stream.

    The Amazon Web Services KMS key identifier is the key ARN, key ID, alias
    ARN, or alias name for the KMS key.
    """
    ActivityStreamKinesisStreamName: str = Field(default=None, frozen=True)
    """
    The name of the Amazon Kinesis data stream used for the database activity
    stream.
    """
    ActivityStreamMode: Literal["sync", "async"] = Field(default=None, frozen=True)
    """
    The mode of the database activity stream.

    Database events such as a change or access generate an activity stream
    event. RDS for Oracle always handles these events asynchronously.
    """
    ActivityStreamEngineNativeAuditFieldsIncluded: bool = Field(
        default=None, frozen=True
    )
    """
    Indicates whether engine-native audit fields are included in the database
    activity stream.
    """
    AutomationMode: Optional[Literal["full", "all-paused"]] = None
    """
The automation mode of the RDS Custom DB instance: ``full`` or ``all paused``.
If ``full``, the DB instance automates monitoring and instance recovery. If
``all paused``, the instance pauses automation for the duration set by
``--resume-full-automation-mode-minutes``.
    """
    ResumeFullAutomationModeTime: datetime = Field(default=None, frozen=True)
    """
    The number of minutes to pause the automation.

    When the time period ends, RDS Custom resumes full automation. The minimum
    value is 60 (default). The maximum value is 1,440.
    """
    CustomIamInstanceProfile: Optional[str] = None
    """
    The instance profile associated with the underlying Amazon EC2 instance of
    an RDS Custom DB instance.

    The instance profile must meet the following
    requirements:
    """
    BackupTarget: Optional[str] = None
    """
The location where automated backups and manual snapshots are stored: Amazon
Web Services Outposts or the Amazon Web Services Region.
    """
    NetworkType: Optional[str] = None
    """
    The network type of the DB instance.
    """
    ActivityStreamPolicyStatus: Literal[
        "locked", "unlocked", "locking-policy", "unlocking-policy"
    ] = Field(default=None, frozen=True)
    """
    The status of the policy state of the activity stream.
    """
    StorageThroughput: Optional[int] = None
    """
    The storage throughput for the DB instance.
    """
    DBSystemId: Optional[str] = None
    """
    The Oracle system ID (Oracle SID) for a container database (CDB).

    The Oracle SID is also the name of the CDB. This setting is only valid for
    RDS Custom DB instances.
    """
    MasterUserSecret: RDSMasterUserSecret = Field(default=None, frozen=True)
    """
    The secret managed by RDS in Amazon Web Services Secrets Manager for the
    master user password.
    """
    CertificateDetails: RDSCertificateDetails = Field(default=None, frozen=True)
    """
    The details of the DB instance's server certificate.
    """
    ReadReplicaSourceDBClusterIdentifier: str = Field(default=None, frozen=True)
    """
    The identifier of the source DB cluster if this DB instance is a read
    replica.
    """
    PercentProgress: str = Field(default=None, frozen=True)
    """
    The progress of the storage optimization operation as a percentage.
    """
    DedicatedLogVolume: Optional[bool] = None
    """
    Indicates whether the DB instance has a dedicated log volume (DLV) enabled.
    """
    IsStorageConfigUpgradeAvailable: bool = Field(default=None, frozen=True)
    """
    Indicates whether an upgrade is recommended for the storage file system
    configuration on the DB instance.

    To migrate to the preferred configuration,
    you can either create a blue/green deployment, or create a read replica from
    the DB instance. For more information, see `Upgrading the storage file system
    for a DB instance <https://docs.aws.amazon.com/AmazonRDS/latest/UserGuide/USER_
    PIOPS.StorageTypes.html#USER_PIOPS.UpgradeFileSystem>`_.
    """
    MultiTenant: Optional[bool] = None
    """
    Specifies whether the DB instance is in the multi-tenant configuration
    (TRUE) or the single-tenant configuration (FALSE).
    """
    EngineLifecycleSupport: Optional[str] = None
    """
    The life cycle type for the DB instance.
    """

    @property
    def pk(self) -> Optional[str]:
        """
        Return the primary key of the model.   This is the value of the
        :py:attr:`DBInstanceArn` attribute.

        Returns:
            The primary key of the model instance.
        """
        return self.DBInstanceArn

    @property
    def arn(self) -> Optional[str]:
        """
        Return the ARN of the model.   This is the value of the
        :py:attr:`DBInstanceArn` attribute.

        Returns:
            The ARN of the model instance.
        """
        return self.DBInstanceArn

    @property
    def name(self) -> Optional[str]:
        """
        Return the name of the model.   This is the value of the
        :py:attr:`DBInstanceIdentifier` attribute.

        Returns:
            The name of the model instance.
        """
        return self.DBInstanceIdentifier


# =======================
# Request/Response Models
# =======================


class CreateDBInstanceResult(Boto3Model):
    RDSDBInstance: DBInstance = Field(default=None, serialization_alias="DBInstance")
    """
    Contains the details of an Amazon RDS DB instance.
    """


class CloudwatchLogsExportConfiguration(Boto3Model):
    """
    The log types to be enabled for export to CloudWatch Logs for a specific DB
    instance.

    A change to the ``CloudwatchLogsExportConfiguration`` parameter is always
    applied to the DB instance immediately. Therefore, the ``ApplyImmediately``
    parameter has no effect.

    This setting doesn't apply to RDS Custom DB instances.
    """

    EnableLogTypes: Optional[List[str]] = None
    """
    The list of log types to enable.
    """
    DisableLogTypes: Optional[List[str]] = None
    """
    The list of log types to disable.
    """


class ModifyDBInstanceResult(Boto3Model):
    RDSDBInstance: DBInstance = Field(default=None, serialization_alias="DBInstance")
    """
    Contains the details of an Amazon RDS DB instance.
    """


class DeleteDBInstanceResult(Boto3Model):
    RDSDBInstance: DBInstance = Field(default=None, serialization_alias="DBInstance")
    """
    Contains the details of an Amazon RDS DB instance.
    """


class DBInstanceMessage(Boto3Model):
    """
    Contains the result of a successful invocation of the
    ``DescribeDBInstances`` action.
    """

    Marker: Optional[str] = None
    """
    An optional pagination token provided by a previous request.

    If this parameter
    is specified, the response includes only records beyond the marker, up to the
    value specified by ``MaxRecords`` .
    """
    DBInstances: Optional[List["DBInstance"]] = None
    """
    A list of ``DBInstance`` instances.
    """



# This file is automatically generated by botocraft.  Do not edit directly.
# pylint: disable=anomalous-backslash-in-string,unsubscriptable-object,line-too-long,arguments-differ,arguments-renamed,unused-import,redefined-outer-name
# pyright: reportUnusedImport=false
# mypy: disable-error-code="index, override, assignment, union-attr, misc"
from datetime import datetime
from typing import Any, ClassVar, Dict, List, Literal, Optional, Type, cast

from pydantic import Field

from botocraft.mixins.tags import TagsDictMixin
from botocraft.services.common import Tag

from .abstract import (Boto3Model, Boto3ModelManager, PrimaryBoto3Model,
                       ReadonlyBoto3Model, ReadonlyBoto3ModelManager,
                       ReadonlyPrimaryBoto3Model)

# ===============
# Managers
# ===============


class ParameterManager(Boto3ModelManager):

    service_name: str = "ssm"

    def create(
        self,
        model: "Parameter",
        Description: Optional[str] = None,
        KeyId: Optional[str] = None,
        AllowedPattern: Optional[str] = None,
        Tags: Optional[List[Tag]] = None,
        Tier: Optional[Literal["Standard", "Advanced", "Intelligent-Tiering"]] = None,
        Policies: Optional[str] = None,
    ) -> int:
        """
        Add a parameter to the system.

        Args:
            model: The :py:class:``Parameter`` to create.

        Keyword Args:
            Description: Information about the parameter that you want to add to the
                system. Optional but recommended.
            KeyId: The Key Management Service (KMS) ID that you want to use to encrypt
                a parameter. Use a custom key for better security. Required for parameters
                that use the ``SecureString`` data type.
            AllowedPattern: A regular expression used to validate the parameter value.
                For example, for String types with values restricted to numbers, you can
                specify the following: AllowedPattern=^d+$
            Tags: Optional metadata that you assign to a resource. Tags enable you to
                categorize a resource in different ways, such as by purpose, owner, or
                environment. For example, you might want to tag a Systems Manager parameter
                to identify the type of resource to which it applies, the environment, or
                the type of configuration data referenced by the parameter. In this case,
                you could specify the following key-value pairs:
            Tier: The parameter tier to assign to a parameter.
            Policies: One or more policies to apply to a parameter. This operation
                takes a JSON array. Parameter Store, a capability of Amazon Web Services
                Systems Manager supports the following policy types:
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            Name=data.get("Name"),
            Value=data.get("Value"),
            Description=self.serialize(Description),
            Type=data.get("Type"),
            KeyId=self.serialize(KeyId),
            Overwrite=data.get("Overwrite"),
            AllowedPattern=self.serialize(AllowedPattern),
            Tags=self.serialize(Tags),
            Tier=self.serialize(Tier),
            Policies=self.serialize(Policies),
            DataType=data.get("DataType"),
        )
        _response = self.client.put_parameter(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = PutParameterResult(**_response)

        self.sessionize(response.Version)
        return cast(int, response.Version)

    def update(
        self,
        model: "Parameter",
        Description: Optional[str] = None,
        KeyId: Optional[str] = None,
        Overwrite: Optional[bool] = None,
        AllowedPattern: Optional[str] = None,
        Tags: Optional[List[Tag]] = None,
        Tier: Optional[Literal["Standard", "Advanced", "Intelligent-Tiering"]] = None,
        Policies: Optional[str] = None,
    ) -> int:
        """
        Add a parameter to the system.

        Args:
            model: The :py:class:``Parameter`` to update.

        Keyword Args:
            Description: Information about the parameter that you want to add to the
                system. Optional but recommended.
            KeyId: The Key Management Service (KMS) ID that you want to use to encrypt
                a parameter. Use a custom key for better security. Required for parameters
                that use the ``SecureString`` data type.
            Overwrite: Overwrite an existing parameter. The default value is ``false``.
            AllowedPattern: A regular expression used to validate the parameter value.
                For example, for String types with values restricted to numbers, you can
                specify the following: AllowedPattern=^d+$
            Tags: Optional metadata that you assign to a resource. Tags enable you to
                categorize a resource in different ways, such as by purpose, owner, or
                environment. For example, you might want to tag a Systems Manager parameter
                to identify the type of resource to which it applies, the environment, or
                the type of configuration data referenced by the parameter. In this case,
                you could specify the following key-value pairs:
            Tier: The parameter tier to assign to a parameter.
            Policies: One or more policies to apply to a parameter. This operation
                takes a JSON array. Parameter Store, a capability of Amazon Web Services
                Systems Manager supports the following policy types:
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            Name=data.get("Name"),
            Value=data.get("Value"),
            Description=self.serialize(Description),
            Type=data.get("Type"),
            KeyId=self.serialize(KeyId),
            Overwrite=self.serialize(Overwrite),
            AllowedPattern=self.serialize(AllowedPattern),
            Tags=self.serialize(Tags),
            Tier=self.serialize(Tier),
            Policies=self.serialize(Policies),
            DataType=data.get("DataType"),
        )
        _response = self.client.put_parameter(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = PutParameterResult(**_response)

        self.sessionize(response.Version)
        return cast(int, response.Version)

    def get(self, Name: str, *, WithDecryption: bool = True) -> Optional["Parameter"]:
        """
        Get information about one or more parameters by specifying multiple
        parameter names.

        Args:
            Name: The name of the parameter you want to query.

        Keyword Args:
            WithDecryption: Return decrypted secure string value. Return decrypted
                values for secure string parameters. This flag is ignored for ``String``
                and ``StringList`` parameter types.
        """
        args: Dict[str, Any] = dict(
            Names=self.serialize([Name]), WithDecryption=self.serialize(WithDecryption)
        )
        _response = self.client.get_parameters(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = GetParametersResult(**_response)

        if response and response.Parameters:
            self.sessionize(response.Parameters[0])
            return response.Parameters[0]
        return None

    def get_many(
        self, Names: List[str], *, WithDecryption: bool = True
    ) -> List["Parameter"]:
        """
        Get information about one or more parameters by specifying multiple
        parameter names.

        Args:
            Names: The names or Amazon Resource Names (ARNs) of the parameters that you
                want to query. For parameters shared with you from another account, you
                must use the full ARNs.

        Keyword Args:
            WithDecryption: Return decrypted secure string value. Return decrypted
                values for secure string parameters. This flag is ignored for ``String``
                and ``StringList`` parameter types.
        """
        args: Dict[str, Any] = dict(
            Names=self.serialize(Names), WithDecryption=self.serialize(WithDecryption)
        )
        _response = self.client.get_parameters(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = GetParametersResult(**_response)

        self.sessionize(respsonse.Parameters)
        return response.Parametersj

    def list(
        self,
        *,
        Filters: Optional[List["ParametersFilter"]] = None,
        ParameterFilters: Optional[List["ParameterStringFilter"]] = None,
        Shared: Optional[bool] = None
    ) -> List["ParameterMetadata"]:
        """
        Lists the parameters in your Amazon Web Services account or the parameters
        shared with you when you enable the
        `Shared <https://docs.aws.amazon.com/systems-
        manager/latest/APIReference/API_DescribeParameters.html#systemsmanager-
        DescribeParameters-request-Shared>`_ option.

        Keyword Args:
            Filters: This data type is deprecated. Instead, use ``ParameterFilters``.
            ParameterFilters: Filters to limit the request results.
            Shared: Lists parameters that are shared with you.

        """
        paginator = self.client.get_paginator("describe_parameters")
        args: Dict[str, Any] = dict(
            Filters=self.serialize(Filters),
            ParameterFilters=self.serialize(ParameterFilters),
            Shared=self.serialize(Shared),
        )
        response_iterator = paginator.paginate(
            **{k: v for k, v in args.items() if v is not None}
        )
        results: List["ParameterMetadata"] = []
        for _response in response_iterator:
            response = DescribeParametersResult(**_response)
            if response.Parameters:
                results.extend(response.Parameters)
            else:
                break
        self.sessionize(results)
        return results

    def delete(self, Name: str) -> None:
        """
        Delete a parameter from the system. After deleting a parameter, wait
        for at least 30 seconds to create a parameter with the same name.

        Args:
            Name: The name of the parameter to delete.
        """
        args: Dict[str, Any] = dict(Name=self.serialize(Name))
        self.client.delete_parameter(**{k: v for k, v in args.items() if v is not None})


# ==============
# Service Models
# ==============


class Parameter(PrimaryBoto3Model):
    """
    An Amazon Web Services Systems Manager parameter in Parameter Store.
    """

    manager_class: ClassVar[Type[Boto3ModelManager]] = ParameterManager

    Name: str
    """
    The name of the parameter.
    """
    Value: Optional[str] = None
    """
    The parameter value.
    """
    Type: Literal["String", "StringList", "SecureString"]
    """
    The type of parameter.

    Valid values include the following: ``String``,
    ``StringList``, and ``SecureString``.
    """
    DataType: Optional[str] = "text"
    """
    The data type of the parameter, such as ``text`` or ``aws:ec2:image``.

    The
    default is ``text``.
    """
    Version: int = Field(default=None, frozen=True)
    """
    The parameter version.
    """
    Selector: str = Field(default=None, frozen=True)
    """
    Either the version number or the label used to retrieve the parameter
    value.

    Specify selectors by using one of the following formats:
    """
    SourceResult: str = Field(default=None, frozen=True)
    """
    Applies to parameters that reference information in other Amazon Web
    Services services.

    ``SourceResult`` is the raw result or response from the source.
    """
    LastModifiedDate: datetime = Field(default=None, frozen=True)
    """
    Date the parameter was last changed or updated and the parameter version
    was created.
    """
    ARN: str = Field(default=None, frozen=True)
    """
    The Amazon Resource Name (ARN) of the parameter.
    """

    @property
    def pk(self) -> Optional[str]:
        """
        Return the primary key of the model.   This is the value of the
        :py:attr:`Name` attribute.

        Returns:
            The primary key of the model instance.
        """
        return self.Name

    @property
    def arn(self) -> Optional[str]:
        """
        Return the ARN of the model.   This is the value of the :py:attr:`ARN`
        attribute.

        Returns:
            The ARN of the model instance.
        """
        return self.ARN

    @property
    def name(self) -> Optional[str]:
        """
        Return the name of the model.   This is the value of the
        :py:attr:`Name` attribute.

        Returns:
            The name of the model instance.
        """
        return self.Name


# =======================
# Request/Response Models
# =======================


class PutParameterResult(Boto3Model):
    Version: Optional[int] = None
    """
    The new version number of a parameter.

    If you edit a parameter value, Parameter Store automatically creates a new
    version and assigns this new version a unique ID. You can reference a
    parameter version ID in API operations or in Systems Manager documents (SSM
    documents). By default, if you don't specify a specific version, the system
    returns the latest parameter value when a parameter is called.
    """
    Tier: Optional[Literal["Standard", "Advanced", "Intelligent-Tiering"]] = None
    """
    The tier assigned to the parameter.
    """


class GetParametersResult(Boto3Model):
    Parameters: Optional[List["Parameter"]] = None
    """
    A list of details for a parameter.
    """
    InvalidParameters: Optional[List[str]] = None
    """
    A list of parameters that aren't formatted correctly or don't run during an
    execution.
    """


class ParametersFilter(Boto3Model):
    """
    This data type is deprecated.

    Instead, use ParameterStringFilter.
    """

    Key: Literal["Name", "Type", "KeyId"]
    """
    The name of the filter.
    """
    Values: List[str]
    """
    The filter values.
    """


class ParameterStringFilter(Boto3Model):
    """
    One or more filters.

    Use a filter to return a more specific list of results.
    """

    Key: str
    """
    The name of the filter.
    """
    Option: Optional[str] = None
    """
    For all filters used with DescribeParameters, valid options include
    ``Equals`` and ``BeginsWith``.

    The ``Name`` filter additionally supports the ``Contains``
    option. (Exception: For filters using the key ``Path``, valid options include
    ``Recursive`` and ``OneLevel``.)
    """
    Values: Optional[List[str]] = None
    """
    The value you want to search for.
    """


class ParameterInlinePolicy(Boto3Model):
    """
    One or more policies assigned to a parameter.
    """

    PolicyText: Optional[str] = None
    """
    The JSON text of the policy.
    """
    PolicyType: Optional[str] = None
    """
    The type of policy.

    Parameter Store, a capability of Amazon Web Services Systems Manager,
    supports the following policy types: Expiration, ExpirationNotification,
    and NoChangeNotification.
    """
    PolicyStatus: Optional[str] = None
    """
    The status of the policy.

    Policies report the following statuses: Pending (the
    policy hasn't been enforced or applied yet), Finished (the policy was applied),
    Failed (the policy wasn't applied), or InProgress (the policy is being applied
    now).
    """


class ParameterMetadata(Boto3Model):
    """
    Metadata includes information like the Amazon Resource Name (ARN) of the
    last user to update the parameter and the date and time the parameter was
    last used.
    """

    Name: Optional[str] = None
    """
    The parameter name.
    """
    ARN: Optional[str] = None
    """
    The (ARN) of the last user to update the parameter.
    """
    Type: Optional[Literal["String", "StringList", "SecureString"]] = None
    """
    The type of parameter.

    Valid parameter types include the following: ``String``,
    ``StringList``, and ``SecureString``.
    """
    KeyId: Optional[str] = None
    """
    The alias of the Key Management Service (KMS) key used to encrypt the
    parameter.

    Applies to ``SecureString`` parameters only.
    """
    LastModifiedDate: Optional[datetime] = None
    """
    Date the parameter was last changed or updated.
    """
    LastModifiedUser: Optional[str] = None
    """
    Amazon Resource Name (ARN) of the Amazon Web Services user who last changed
    the parameter.
    """
    Description: Optional[str] = None
    """
    Description of the parameter actions.
    """
    AllowedPattern: Optional[str] = None
    """
    A parameter name can include only the following letters and symbols.
    """
    Version: Optional[int] = None
    """
    The parameter version.
    """
    Tier: Optional[Literal["Standard", "Advanced", "Intelligent-Tiering"]] = None
    """
    The parameter tier.
    """
    Policies: Optional[List["ParameterInlinePolicy"]] = None
    """
    A list of policies associated with a parameter.
    """
    DataType: Optional[str] = None
    """
    The data type of the parameter, such as ``text`` or ``aws:ec2:image``.

    The
    default is ``text``.
    """


class DescribeParametersResult(Boto3Model):
    Parameters: Optional[List["ParameterMetadata"]] = None
    """
    Parameters returned by the request.
    """
    NextToken: Optional[str] = None
    """
    The token to use when requesting the next set of items.
    """


class DeleteParameterResult(Boto3Model):
    pass
